/*
===============================================================================

  FILE:  SVreader_svg.h

  CONTENTS:

    Reads a Streaming Volume Mesh from a binary format.
    Handles Tet and hex mixed meshes with arbitrary number of vertex and cell properties.
    The format is little endian.

    HEADER:
	unsigned char		SVG version
	int32			number of vertices
	int32			number of cells
	int32			number of vertex properties (including geometry)
	int32			number of cell properties (including geometry)
	v_pnum_f x float	vertex properties bbox max
	v_pnum_f x float	vertex properties bbox min
	c_pnum_f x float	cell properties bbox max
	c_pnum_f x float	cell properties bbox min
	int32			Must be zero (reserved for future use).
	int32			extra header data size. Must be zero (reserved for future use).

    DATA:
	{
		uint8				element type
						(currrently SVG_VERTEX, SVG_TET and SVG_HEX are supported)
		#SVG_VERTEX
			v_pnum_f x float	vertex properties
		#SVG_TET
			4 x int32		vertex indices
			c_pnum_f x float	cell properties
		#SVG_HEX
			8 x int32		vertex indices
			c_pnum_f x float	cell properties
	} repeated nverts+ncells times


  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009 clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    6 November 2009 -- Created

===============================================================================
*/
#ifndef SVREADER_SVG_H
#define SVREADER_SVG_H

#include "svreader.h"

#include <stdio.h>

class SVreader_svg : public SVreader
{
	public:

		// SVreader interface function implementations

		SVevent read_element();
		SVevent read_event();

		void close ( bool close_file=true );

		// SVreader_svg functions

		bool open ( FILE* fp );

		SVreader_svg();
		~SVreader_svg();

	private:
		FILE* file;
		int have_finalized, next_finalized;
		I32* finalized_vertices;

		void read_header();
		void read_buffer();

		inline void readInt32 ( I32 *i );
		inline void readUChar ( unsigned char *c );
		inline void readFloat ( float *f );
		inline void readNFloats ( float *f, int n );
};

#endif
