/*
===============================================================================

  FILE:  geo_structures.h

  CONTENTS:

    Implementation of GMSH structures


  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009 clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    23 November 2009 -- Created

===============================================================================
*/
#ifndef GEOSTRUCTURES_H
#define GEOSTRUCTURES_H

#include <math.h>
#include <assert.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
  #include <ext/hash_map>
  namespace std { using namespace __gnu_cxx; }
  using namespace std;
#else
  #include <hash_map.h>
  #include <list.h>
  #include <vector.h>
#endif

//#define OPTIMIZE_CACHE_MISS

namespace Geo
{
  typedef list< pair<int, float*> *> StreamList;

  class Coords
  {
    public:
      int x;
      int y;
      int z;
  };

  inline float dot ( float *u, float*v )
  {
    return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
  }

  inline void matMul ( const float *A, float*v )
  {
    float tv[3];
    tv[0] =  A[0]*v[0] + A[1]*v[1] + A[2]*v[2];
    tv[1] =  A[3]*v[0] + A[4]*v[1] + A[5]*v[2];
    tv[2] =  A[6]*v[0] + A[7]*v[1] + A[8]*v[2];
    v[0] = tv[0];
    v[1] = tv[1];
    v[2] = tv[2];
  }

  class Point
  {
    public:
      enum PointType {TYPE_CORNER, TYPE_ON_CURVE, TYPE_ON_SURFACE, TYPE_INTERIOR};

      Point(int i, PointType t, int nSlots): geo_index(i), idx_index(-1), type(t), nAvailableSlots(nSlots)
      {}

      inline int ind() const { return geo_index; }

      inline int idx() const { return idx_index; }

      inline void set_idx(int i) { idx_index = i; }

      inline bool streamed() const { return idx_index!=-1; } //has the point already been streamed ?

      inline bool isInterior() const { return type==TYPE_INTERIOR; }

      inline PointType getType() const { return type; }

      inline bool hasAvailableSlots() { return nAvailableSlots>0; }

      inline void takeSlot() { assert(nAvailableSlots); --nAvailableSlots; }

      inline void addSlot() { ++nAvailableSlots; }

      float pos[3];
      
    private:
      int geo_index; //this is the index of the point in the geo file
      int idx_index; //this is the index of the vertex in the mesh
      PointType type;
      int nAvailableSlots; //number of hexahedron that can still reference this vertex. When this becomes null, the vertex should be finalized and deleted.
  };

  
  class CyclicPointArray4
  {
    public:
      inline Point*& operator[] (int i)
      {
        if(i<0) return array[((-i/4) + 1)*4];
        else return array[i%4];
      }

#ifdef DEBUG
      friend std::ostream& operator<<(std::ostream& out, const CyclicPointArray4& object)
      {
        for(int i=0; i<4; ++i)
          out << object.array[i]->ind() << ", ";
        out << std::endl;
        return out;
      }
#endif

    private:
      Point* array[4];
  };

  class Curve
  {
    public:
      Curve(int i, Point *s, Point *e)
      :geo_index(i), subdiv(0), start(s), end(e), nSurfaces(0), visited(false)
      {
      }
      virtual ~Curve(){}

      //get strat and end points
      inline Point* startPoint(){ return start; }

      inline Point* endPoint(){ return end; }

      //mesh the curve by subdividing. v_count is the base for indices (i.e. the new vertices will be numbered from there on, without using this number, and will increment it. )
      virtual void generateMesh() = 0;

      inline Point* operator () (int i) const;

      //set the number of subdivisions
      inline void setSubdiv(int s) { subdiv = s; }

      inline int getSubdiv() const { return subdiv; }

      inline int ind() const { return geo_index; }

      inline bool isVisited() const { return visited; };

      inline void addSurface() ;

    protected:
      int geo_index;
      int subdiv;
      std::vector<Point*> mesh;
      Point *start;
      Point *end;
      int nSurfaces; //the number of block surfaces that use this curve
      bool visited;

      inline void setVisited() { visited = true; };
  };

  class Line:public Curve
  {
    public:
      Line ( int index, Point *s, Point *e ) : Curve(index, s, e) {};
      virtual ~Line(){}

      void generateMesh();
    private:
      inline void getAtI ( int i, float *p );
  };

  class Circle:public Curve
  {
    public:
      Circle ( int index, Point *s, Point *c, Point *e );
      virtual ~Circle(){}

      void generateMesh();

    private:
      Point *center;
      float R2;
      float theta;

  };

  class Block;

  class Surface
  {
    public:
      //build a surface with index i from the 4-curves loop cp. cp will be deleted on surface destruction, no it should not be deleted by the application. if r[k] is true, the curve cp[k] is taken in the reverse orientation.
      Surface ( int i, Curve **cp, bool *r ) ;

      //add a block to the surface
      inline void addBlock(Block* b);

      inline Block* getBlock(int bNum);
      
      ~Surface();

      //get the point at (x,y)
      Point* operator ()(int x, int y);

      //get the four corner vertices of the block in the order of curves startPoints
      void getCorners ( CyclicPointArray4 &s_idx, bool reverseNormal );
      void getCorners ( Point** s_idx, bool reverseNormal );

      //careful! the returned normal is not scaled...
      void getCenterAndNormal(float *center, float * normal);

      void generateMesh();

      void getSubdiv(int &subdivX, int &subdivY);

      inline bool isVisited() const { return visited; };
  
      inline int ind() const { return geo_index; }

    private:
      int geo_index;
      Curve ** curves;
      bool *reverse;
      Block *blocks[2]; //a surface can only be shared by two blocks in the manifold case
      bool visited;
      std::vector<Point*> mesh;

      inline void setVisited() { visited = true;};
  };

  class Block
  {
    //typedef hash_map< Coords<int> , Point*> CoordsPointHash;
    typedef hash_map< int , Point*> CoordsPointHash;
    public:
      //builds a block from surfaces sp. The faces are first reordered to fit the SVH orientation. The faces orientation of the faces is computed and put in <reverse> such that if if reverse[k] is false, then surface[k] has its curves enumerated in right-hand, inwards pointing order. Else, surface[k] is in right-hand, outwards pointing order. sp has to be destroyed after a call to this function.
      Block ( int i, Surface **sp ) ;

      ~Block();

      //get the eight corner vertices of the block in the CUBIT order.
      void getCorners(Point **corners);

      //get the eight corner vertices of the block in the CUBIT order (slow version that does not need sRotation)
      void getCornersSlow(Point **corners);

      void beginMeshing();

      bool getNextHex(int *c_idx, StreamList &verticesToStream, bool *final, int &v_count);

      void endMeshing();

      inline int ind() const { return geo_index; }

    private:
      int geo_index;
      Surface* surfaces[6]; //the surfaces that define the block, in SVH order.
      bool reverse[6]; //the orientation of each surface. false means normal pointing inwards
      int sRotation[6]; //the rotation to apply to each face so that the startPoint of its first Curve (resp. the endpoint of its last curve if reverse is true) is in 0 for f0, f1 and f2, v1 for f3, v3 for f4, and v4 for f5.

      //for meshing:
      bool meshing;
      Coords currentHexCoords; //(x,y,z) for current Hexahedron
      CoordsPointHash *activeVerts;
      int nHexas;
      int subdivX, subdivY, subdivZ;
#ifdef OPTIMIZE_CACHE_MISS
      bool traversalDirX, traversalDirY;
#endif

      //get the vertex at scanline coordinates vXc. If the vertex is not in the hash, create it, put it in the hash. If the vertex has not been stremed yet, andd it to the list verticesToStream. Also return the hash iterator in vXit.
      //inline Point *getPointAtCoords(Coords<int> &vXc, std::list<float*> &verticesToStream, bool &hasBeenCreated, CoordsPointHash::iterator &vXit, int &v_count)
      void getPointAtCoords(int vXc, StreamList &verticesToStream, CoordsPointHash::iterator &vXit, int &v_count);

      //tranfinite interpolation for the given point at coords vXc
      void transfiniteInterpolation(int vXc, float *pos );

      void getLocalSurfaceCoords(int surfaceIndex, Coords &blockCoords, int &localX, int &localY);
  };

}

#endif
