/*
===============================================================================

  FILE:  svreader_hzip.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2005  clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_hzip.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "vecnfv.h"
#include "vecniv.h"

#include <string.h>
#include <assert.h>

#define HZIP_CUBIT_PERMUTATION 0x67542310

//#define PRINT_DEBUG_OUTPUT

//#define READ_COORDS_FROM_FILE
#ifdef READ_COORDS_FROM_FILE
FILE * iCoordFile=NULL;
#endif

bool SVreader_hzip::open ( FILE* file )
{
#ifndef MULTIPLEX
	fprintf ( stderr, "WARNING: SVreader_hzip ignores the file argument, and reads the connectivity and geometry in hzip_conn and hzip_geo respectively.\n" );
#endif

	if ( file == 0 )
	{
		return false;
	}

#ifdef _WIN32
	if ( file == stdin )
	{
		if ( _setmode ( _fileno ( stdin ), _O_TEXT ) == -1 )
		{
			fprintf ( stderr, "ERROR: cannot set stdin to text (translated) mode\n" );
		}
	}
#endif

	this->file = file;

	v_count = 0;
	c_count = 0;

#ifdef MULTIPLEX
	conn_stream = hzip_mesh_open_file ( file, HZIP_CUBIT_PERMUTATION );
	HZNstream *tgs = hzip_node_open_file ( file, HZIP_CUBIT_PERMUTATION, 0 );
	v_pnum_f = hzip_node_dimensions(tgs);
	nverts = hzip_node_count(tgs);
		
	geo_stream = new HZNstream*[v_pnum_f];
	geo_stream[0] = tgs;

	for(int i=1; i<v_pnum_f; ++i)
	{
	// 	if ( v_pmin_f && v_pmax_f )
	// 		geo_stream = hzip_node_create_file ( geo_file, HZIP_CUBIT_PERMUTATION, 0, 3, nverts==-1 ? 0 : nverts, hzINT, HZM_CODEC_ZLIB, 0 );
	// 	else
			geo_stream[i] = hzip_node_open_file ( file, HZIP_CUBIT_PERMUTATION, 0 );
	}
#else
	conn_file = fopen ( "hzip_conn", "rb" );
	if ( !conn_file )
	{
		fprintf ( stderr, "ERROR: Cannot open hzip_conn for writing.\n" );
		return false;
	}
	conn_stream = hzip_mesh_open_file ( conn_file, HZIP_CUBIT_PERMUTATION );

	FILE* tfile = fopen ( "hzip_geo_0", "rb" );
	if(!tfile)
	{
		v_pnum_f = 0;
		nverts=-1;
	}
	else
	{
		HZNstream *tgs = hzip_node_open_file ( tfile, HZIP_CUBIT_PERMUTATION, 0 );
		v_pnum_f = hzip_node_dimensions(tgs);
		nverts = hzip_node_count(tgs);
		
		geo_file = new FILE*[v_pnum_f];
		geo_stream = new HZNstream*[v_pnum_f];
		geo_file[0] = tfile;
		geo_stream[0] = tgs;

		char buf[1024];
		for(int i=1; i<v_pnum_f; ++i)
		{
			sprintf(buf, "hzip_geo_%i", i);
			geo_file[i] = fopen ( buf, "rb" );
			if ( !geo_file )
			{
				fprintf ( stderr, "ERROR: Cannot open %s for writing.\n", buf );
				return false;
			}
		// 	if ( v_pmin_f && v_pmax_f )
		// 		geo_stream = hzip_node_create_file ( geo_file, HZIP_CUBIT_PERMUTATION, 0, 3, nverts==-1 ? 0 : nverts, hzINT, HZM_CODEC_ZLIB, 0 );
		// 	else
				geo_stream[i] = hzip_node_open_file ( geo_file[i], HZIP_CUBIT_PERMUTATION, 0 );
		}
	}

#endif //MULTIPLEX

	c_pnum_f = 0;
	if ( v_pnum_f ) v_prop_f = new float[v_pnum_f];
	if ( c_pnum_f ) c_prop_f = new float[c_pnum_f];

	if(next_props) delete[] next_props;
	next_props = new float[v_pnum_f*8];

	c_idx = new I32[8];
	c_final = new bool[8];
	for(int i=0; i<8; ++i) c_final[i] = false;

	ncells = hzip_mesh_cells(conn_stream);

	if(ncells==0 || nverts==0)
	{
		fprintf(stderr, "ERROR: ncells=%i and nverts=%i\n", ncells, nverts);
		return false;
	}

	if(vertexDecoded) delete vertexDecoded;
	vertexDecoded = new std::vector<bool>(nverts, false);

	if(vertices) delete[] vertices;
	vertices = new float[nverts*v_pnum_f];


#ifdef READ_COORDS_FROM_FILE
	iCoordFile = fopen("coords", "rb");
#endif

	//read first cell
	hasNextCell = readNextCell();

	return true;
}

void SVreader_hzip::close ( bool close_file )
{
	// close of SVreader_hzip
#ifdef MULTIPLEX
	hzip_mesh_close ( conn_stream );
	for(int i=0; i<v_pnum_f; ++i)
	{
		hzip_node_close ( geo_stream[i] );
	}
#else
	hzip_mesh_close ( conn_stream );
	fclose ( conn_file );
	for(int i=0; i<v_pnum_f; ++i)
	{
		hzip_node_close ( geo_stream[i] );
		fclose ( geo_file[i] );
	}
	delete[] geo_file;
	delete[] geo_stream;
#endif

	// close of SVreader interface
	v_count = -1;
	c_count = -1;
	if ( v_prop_f ) delete [] v_prop_f;
	v_prop_f = 0;
	if ( c_idx ) delete [] c_idx;
	c_idx = 0;
	if ( c_final ) delete [] c_final;
	c_final = 0;

#ifdef READ_COORDS_FROM_FILE
	fclose(iCoordFile);
#endif
}

bool SVreader_hzip::readNextCell()
{
  int i;
	//read the next cell
	int t = hzip_mesh_read_cell(conn_stream, next_idx);
	if(!t==1) return false;
	//are there some unknown vertices ?
	bool hasUnknownVertices = false;
	for(i=0; i<8; ++i)
	{
		if(!(*vertexDecoded)[next_idx[i]])
		{
			hasUnknownVertices = true;
			(*vertexDecoded)[next_idx[i]] = true;
			decoded[i] = false;
		}
		else
		{
			decoded[i] = true;
			for(int k=0; k<v_pnum_f; ++k)
				next_props[k*8+i] = vertices[next_idx[i]*v_pnum_f+k];
		}
	}
	//if the cell has at least one vertex that needs decoding, decode all the vertices in the cell.
	if(hasUnknownVertices)
	{
#ifdef READ_COORDS_FROM_FILE
		for(i=0; i<8; ++i)
		{
			if(!decoded[i])
			{
				int dummy;
				fscanf(iCoordFile, "%i %f %f %f\n", &dummy, vertices + next_idx[i]*v_pnum_f, vertices + next_idx[i]*v_pnum_f + 1, vertices + next_idx[i]*v_pnum_f + 2);
			}
		}
#else
		bool decoded2[8];
		for(int k=0; k<v_pnum_f; ++k)
		{
			for(i=0; i<8; ++i)
				decoded2[i] = decoded[i];
			hzip_node_read_unknown_cell_vertices( geo_stream[k], next_props+k*8, decoded2 );
			for(i=0; i<8; ++i) assert(decoded2[i]);
		}
		for(i=0; i<8; ++i)
		{
			if(!decoded[i])
				for(int k=0; k<v_pnum_f; ++k)
					vertices[next_idx[i]*v_pnum_f+k] = next_props[k*8+i];
		}
#endif
	}
	return true;
}

SVevent SVreader_hzip::read_element()
{
	if(!hasNextCell)
		return SV_EOF;

	//while there are still unknown vertices in the current cell, read them
	for(int i=0; i<8; ++i)
	{
		if(!decoded[i])
		{
#ifdef PRINT_DEBUG_OUTPUT
			fprintf(stderr, "Vertex %i (index %i) needs streaming.\n", i, next_idx[i]);
#endif
			decoded[i] = true;
			v_idx = next_idx[i];
			memcpy(v_prop_f, vertices+next_idx[i]*v_pnum_f, v_pnum_f*sizeof(float));
			++v_count;
			return SV_VERTEX;
		}
	}
	//when there are no more unknown vertices, output the current cell, and read a new one
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf(stderr, "No more vertices to stream, outputing hex ");
#endif
		for(int i=0; i<8; ++i)
		{
#ifdef PRINT_DEBUG_OUTPUT
			fprintf(stderr, "%i ", next_idx[i]);
#endif
			c_idx[i] = next_idx[i];
		}
#ifdef PRINT_DEBUG_OUTPUT
		fprintf(stderr, "\n\n");
#endif
		++c_count;
		if(c_count<ncells || ncells==-1)
			hasNextCell = readNextCell();
		else
			hasNextCell = false;
		return SV_HEXAHEDRON;
	}
	
	return SV_EOF;
}

SVevent SVreader_hzip::read_event()
{
	return read_element();
}

SVreader_hzip::SVreader_hzip()
{
	// init of SVreader_hzip
	file = 0;
	vertexDecoded = NULL;
	vertices=NULL;
	next_props = NULL;
}

SVreader_hzip::~SVreader_hzip()
{
	// clean-up for SVreader_hzip
}
