/*
===============================================================================

  FILE:  svreader_sva.cpp
  
  CONTENTS:
  
    see corresponding header file
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreader_sva.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "vecnfv.h"
#include "vecniv.h"

#define MAX_LINE_SIZE 512

bool SVreader_sva::open(FILE* file)
{
  if (file == 0)
  {
    return false;
  }

#ifdef _WIN32
  if (file == stdin)
  {
    if(_setmode( _fileno( stdin ), _O_TEXT ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdin to text (translated) mode\n");
    }
  }
#endif

  this->file = file;

  skipped_lines = 0;
  line = (char*)malloc(sizeof(char)*MAX_LINE_SIZE);
  if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
  {
    free(line);
    line = 0;
    return false;
  }

  format = 0;

  // look for header information
  char dummy[MAX_LINE_SIZE];

  while (line && line[0] != 'v' && line[0] != 'c' && line[0] != 't' && line[0] != 'h' && line[0] != 'x')
  {
    if (strstr(line, "svtype"))
    {
      char svtype[MAX_LINE_SIZE];
      sscanf(&(line[1]), "%s %s", dummy, svtype);
      if (strcmp(svtype, "tet") == 0)
      {
        type = SV_TET;
      }
      else if (strcmp(svtype, "hex") == 0)
      {
        type = SV_HEX;
      }
      else if (strcmp(svtype, "mixed") == 0)
      {
        type = SV_MIXED;
      }
      else
      {
        fprintf(stderr,"FATAL ERROR: input type '%s' not supported\n",svtype);
        exit(0);
      }
    }
    else if (strstr(line, "v_pnum_f"))
    {
      sscanf(&(line[1]), "%s %d", dummy, &v_pnum_f);
    }
    else if (strstr(line, "v_pmin_f"))
    {
      if (v_pmin_f) delete v_pmin_f;
      v_pmin_f = new float[v_pnum_f];
      if (v_pnum_f == 3)
      {
        sscanf(&(line[1]), "%s %f %f %f", dummy, &(v_pmin_f[0]), &(v_pmin_f[1]), &(v_pmin_f[2]));
      }
      else if (v_pnum_f == 4)
      {
        sscanf(&(line[1]), "%s %f %f %f %f", dummy, &(v_pmin_f[0]), &(v_pmin_f[1]), &(v_pmin_f[2]), &(v_pmin_f[3]));
      }
      else if (v_pnum_f == 5)
      {
        sscanf(&(line[1]), "%s %f %f %f %f %f", dummy, &(v_pmin_f[0]), &(v_pmin_f[1]), &(v_pmin_f[2]), &(v_pmin_f[3]), &(v_pmin_f[4]));
      }
    }
    else if (strstr(line, "v_pmax_f"))
    {
      if (v_pmax_f) delete v_pmax_f;
      v_pmax_f = new float[v_pnum_f];
      if (v_pnum_f == 3)
      {
        sscanf(&(line[1]), "%s %f %f %f", dummy, &(v_pmax_f[0]), &(v_pmax_f[1]), &(v_pmax_f[2]));
      }
      else if (v_pnum_f == 4)
      {
        sscanf(&(line[1]), "%s %f %f %f %f", dummy, &(v_pmax_f[0]), &(v_pmax_f[1]), &(v_pmax_f[2]), &(v_pmax_f[3]));
      }
      else if (v_pnum_f == 5)
      {
        sscanf(&(line[1]), "%s %f %f %f %f %f", dummy, &(v_pmax_f[0]), &(v_pmax_f[1]), &(v_pmax_f[2]), &(v_pmax_f[3]), &(v_pmax_f[4]));
      }
    }
    else if (strstr(line, "c_pnum_f"))
    {
      sscanf(&(line[1]), "%s %d", dummy, &c_pnum_f);
    }
    else if (strstr(line, "c_pmin_f"))
    {
      if (c_pmin_f) delete c_pmin_f;
      c_pmin_f = new float[c_pnum_f];
      if (c_pnum_f == 1)
      {
        sscanf(&(line[1]), "%s %f", dummy, &(c_pmin_f[0]));
      }
      else if (c_pnum_f == 2)
      {
        sscanf(&(line[1]), "%s %f %f", dummy, &(c_pmin_f[0]), &(c_pmin_f[1]));
      }
      else if (c_pnum_f == 3)
      {
        sscanf(&(line[1]), "%s %f %f %f", dummy, &(c_pmin_f[0]), &(c_pmin_f[1]), &(c_pmin_f[2]));
      }
      else if (c_pnum_f == 4)
      {
        sscanf(&(line[1]), "%s %f %f %f %f", dummy, &(c_pmin_f[0]), &(c_pmin_f[1]), &(c_pmin_f[2]), &(c_pmin_f[3]));
      }
      else if (c_pnum_f == 5)
      {
        sscanf(&(line[1]), "%s %f %f %f %f %f", dummy, &(c_pmin_f[0]), &(c_pmin_f[1]), &(c_pmin_f[2]), &(c_pmin_f[3]), &(c_pmin_f[4]));
      }
    }
    else if (strstr(line, "c_pmax_f"))
    {
      if (c_pmax_f) delete c_pmax_f;
      c_pmax_f = new float[c_pnum_f];
      if (c_pnum_f == 1)
      {
        sscanf(&(line[1]), "%s %f", dummy, &(c_pmax_f[0]));
      }
      else if (c_pnum_f == 2)
      {
        sscanf(&(line[1]), "%s %f %f", dummy, &(c_pmax_f[0]), &(c_pmax_f[1]));
      }
      else if (c_pnum_f == 3)
      {
        sscanf(&(line[1]), "%s %f %f %f", dummy, &(c_pmax_f[0]), &(c_pmax_f[1]), &(c_pmax_f[2]));
      }
      else if (c_pnum_f == 4)
      {
        sscanf(&(line[1]), "%s %f %f %f %f", dummy, &(c_pmax_f[0]), &(c_pmax_f[1]), &(c_pmax_f[2]), &(c_pmax_f[3]));
      }
      else if (c_pnum_f == 5)
      {
        sscanf(&(line[1]), "%s %f %f %f %f %f", dummy, &(c_pmax_f[0]), &(c_pmax_f[1]), &(c_pmax_f[2]), &(c_pmax_f[3]), &(c_pmax_f[4]));
      }
    }
    else if (strstr(line, "bb_min")) // for claudio's *.sma variation
    {
      v_pnum_f = 3;
      if (v_pmin_f) delete v_pmin_f;
      v_pmin_f = new float[v_pnum_f];
      sscanf(&(line[1]), "%s %f %f %f", dummy, &(v_pmin_f[0]), &(v_pmin_f[1]), &(v_pmin_f[2]));
    }
    else if (strstr(line, "bb_max")) // for claudio's *.sma variation
    {
      v_pnum_f = 3;
      if (v_pmax_f) delete v_pmax_f;
      v_pmax_f = new float[v_pnum_f];
      sscanf(&(line[1]), "%s %f %f %f", dummy, &(v_pmax_f[0]), &(v_pmax_f[1]), &(v_pmax_f[2]));
    }
    else if (strstr(line, "nverts"))
    {
      sscanf(&(line[1]), "%s %d", dummy, &nverts);
    }
    else if (strstr(line, "ncells"))
    {
      sscanf(&(line[1]), "%s %d", dummy, &ncells);
    }
    else if (strstr(line, "format") && strstr(line, "CUBIT"))
    {
      format = 0;
    }
    else if (strstr(line, "format") && strstr(line, "GAMBIT"))
    {
      format = 1;
    }
    else if (line[0] == '#')
    {
      //for .hex: try to read two ints which are number of cells and verts.
      int nc, nv;
      if (sscanf(line+1, " %i %i", &nv, &nc) == 2)
      {
        nverts = nv;
        ncells = nc;
      }
      else if (sscanf(&(line[1]), "%s", dummy))
      {
        if (comments == 0)
        {
          ncomments = 0;
          comments = (char**)malloc(sizeof(char*)*10);
          comments[9] = (char*)-1;
        }
        else if (comments[ncomments] == (char*)-1)
        {
          comments = (char**)realloc(comments,sizeof(char*)*ncomments*2);
          comments[ncomments*2-1] = (char*)-1;
        }
        comments[ncomments] = strdup(dummy);
        ncomments++;
      }
    }
    else if (line[0] == 'p' && line[1] == ' ')
    {
      fprintf(stderr,"WARNING: input seems to be a point cloud\n");
    }
    else if (line[0] == 'f' && line[1] == ' ')
    {
      fprintf(stderr,"WARNING: input seems to be a surface mesh\n");
    }
    else
    {
      int j = (int)strlen(line)-1;
      if (j > 0)
      {
        for (int i = 0; i < j; i++)
        {
          if (!isprint(line[i]))
          {
            fprintf(stderr,"WARNING: input file seems to be binary\n");
            break;
          }
        }
        line[(10<j?10:j)] = '\0';
        if (skipped_lines < 5)
        {
          fprintf(stderr,"WARNING: skipping line '%s...'.\n",line);
        }
        else if (skipped_lines == 5)
        {
          fprintf(stderr,"WARNING: skipping more than 5 lines. being quiet.\n");
        }
        skipped_lines++;
      }
    }
    if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
    {
      free(line);
      line = 0;
      return false;
    }
  }
  if (line[0] == 'v')
  {
    post_order = false;
  }
  else if (line[0] == 'c' || line[0] != 't' || line[0] != 'h')
  {
    post_order = true;
  }
  else
  {
    fprintf(stderr,"WARNING: cannot determine order. assuming pre-order SM.\n");
  }

  v_count = 0;
  c_count = 0;

  if (v_pnum_f == 0) v_pnum_f = 3;

  if (v_pnum_f) v_prop_f = new float[v_pnum_f];
  if (c_pnum_f) c_prop_f = new float[c_pnum_f];

  c_idx = new I32[8];
  c_final = new bool[8];
  finalized_vertices = new I32[8]; 

  return true;
}

void SVreader_sva::close(bool close_file)
{
  // close of SVreader_sva
  if (skipped_lines) fprintf(stderr,"WARNING: skipped %d lines.\n",skipped_lines);
  skipped_lines = 0;
  if (close_file && file && file != stdin) fclose(file);
  file = 0;
  if (line)
  {
    free(line);
    line = 0;
  }
  have_finalized = 0; next_finalized = 0;
  delete [] finalized_vertices;
  finalized_vertices = 0;

  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if (v_prop_f) delete [] v_prop_f;
  v_prop_f = 0;
  if (c_idx) delete [] c_idx;
  c_idx = 0;
  if (c_final) delete [] c_final;
  c_final = 0; 
}

SVevent SVreader_sva::read_element()
{
  have_finalized = next_finalized = 0;
  while (line)
  {
    if ((line[0] == 'v') && (line[1] == ' '))
    {
      v_idx = v_count;
      if (v_pnum_f == 3)
      {
        sscanf(&(line[1]), "%f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]));
      }
      else if (v_pnum_f == 4)
      {
        sscanf(&(line[1]), "%f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]));
      }
      else if (v_pnum_f == 5)
      {
        sscanf(&(line[1]), "%f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]));
      }
      else if (v_pnum_f == 6)
      {
        sscanf(&(line[1]), "%f %f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]), &(v_prop_f[5]));
      }
      else if (v_pnum_f == 7)
      {
        sscanf(&(line[1]), "%f %f %f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]), &(v_prop_f[5]), &(v_prop_f[6]));
      }
      else if (v_pnum_f == 8)
      {
        sscanf(&(line[1]), "%f %f %f %f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]), &(v_prop_f[5]), &(v_prop_f[6]), &(v_prop_f[7]));
      }
      else
      {
	fprintf(stderr,"WARNING (SVreader_sva): reading vertices with %d properties not implemented\n", v_pnum_f);
      }
      if (post_order) {finalized_vertices[have_finalized] = v_count; have_finalized++;}
      v_count++;
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
      }
      return SV_VERTEX;
    }
    else if (((line[0] == 'c') && (line[1] == ' ')) || ((line[0] == 't') && (line[1] == ' ')))
    {
      if (c_pnum_f == 0)
      {
        sscanf(&(line[1]), "%d %d %d %d", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]));
      }
      else if (c_pnum_f == 1)
      {
        sscanf(&(line[1]), "%d %d %d %d %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]));
      }
      else if (c_pnum_f == 2)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]));
      }
      else if (c_pnum_f == 3)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]), &(c_prop_f[2]));
      }
      else if (c_pnum_f == 4)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]), &(c_prop_f[2]), &(c_prop_f[3]));
      }
      else if (c_pnum_f == 5)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f %f %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]), &(c_prop_f[2]), &(c_prop_f[3]), &(c_prop_f[4]));
      }
      else if (c_pnum_f == 6)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f %f %f %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]), &(c_prop_f[2]), &(c_prop_f[3]), &(c_prop_f[4]), &(c_prop_f[5]));
      }
      else if (c_pnum_f == 7)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f %f %f %f %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]), &(c_prop_f[2]), &(c_prop_f[3]), &(c_prop_f[4]), &(c_prop_f[5]), &(c_prop_f[6]));
      }
      else if (c_pnum_f == 8)
      {
        sscanf(&(line[1]), "%d %d %d %d %f %f %f %f %f %f %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_prop_f[0]), &(c_prop_f[1]), &(c_prop_f[2]), &(c_prop_f[3]), &(c_prop_f[4]), &(c_prop_f[5]), &(c_prop_f[6]), &(c_prop_f[7]));
      }
      else
      {
        fprintf(stderr,"WARNING (SVreader_sva): reading tetrahedra with %d properties not implemented\n", c_pnum_f);
      }
      c_count++;
      for (int i = 0; i < 4; i++)
      {
        if (c_idx[i] < 0)
        {
          c_idx[i] = v_count+c_idx[i];
          c_final[i] = true;
          finalized_vertices[have_finalized] = c_idx[i];
          have_finalized++;
        }
        else
        {
          c_idx[i] = c_idx[i]-1;
          c_final[i] = false;
        }
      }
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
      }
      c_nverts = 4;
      return SV_TETRAHEDRON;
    }
    else if ((line[0] == 'h') && (line[1] == ' '))
    {
      if (c_pnum_f == 0)
      {
        if (format == 0)
        {
          sscanf(&(line[1]), "%d %d %d %d %d %d %d %d", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_idx[4]), &(c_idx[5]), &(c_idx[6]), &(c_idx[7])); // CUBIT
        }
        else if (format == 1)
        {
          sscanf(&(line[1]), "%d %d %d %d %d %d %d %d", &(c_idx[0]), &(c_idx[1]), &(c_idx[3]), &(c_idx[2]), &(c_idx[4]), &(c_idx[5]), &(c_idx[7]), &(c_idx[6])); // GAMBIT
        }
      }
      else if (c_pnum_f == 1)
      {
        if (format == 0)
        {
          sscanf(&(line[1]), "%d %d %d %d %d %d %d %d %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_idx[4]), &(c_idx[5]), &(c_idx[6]), &(c_idx[7]), &(c_prop_f[0])); // CUBIT
        }
        else if (format == 1)
        {
          sscanf(&(line[1]), "%d %d %d %d %d %d %d %d %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[3]), &(c_idx[2]), &(c_idx[4]), &(c_idx[5]), &(c_idx[7]), &(c_idx[6]), &(c_prop_f[0])); // GAMBIT
        }
      }
      else if (c_pnum_f == 1)
      {
        if (format == 0)
        {
          sscanf(&(line[1]), "%d %d %d %d %d %d %d %d %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]), &(c_idx[4]), &(c_idx[5]), &(c_idx[6]), &(c_idx[7]), &(c_prop_f[0]), &(c_prop_f[1])); // CUBIT
        }
        else if (format == 1)
        {
          sscanf(&(line[1]), "%d %d %d %d %d %d %d %d %f %f", &(c_idx[0]), &(c_idx[1]), &(c_idx[3]), &(c_idx[2]), &(c_idx[4]), &(c_idx[5]), &(c_idx[7]), &(c_idx[6]), &(c_prop_f[0]), &(c_prop_f[1])); // GAMBIT
        }
      }
      else
      {
        fprintf(stderr,"WARNING (SVreader_sva): reading hexahedra with %d properties not implemented\n", c_pnum_f);
      }
      c_count++;
      for (int i = 0; i < 8; i++)
      {
        if (c_idx[i] < 0)
        {
          c_idx[i] = v_count+c_idx[i];
          c_final[i] = true;
          finalized_vertices[have_finalized] = c_idx[i];
          have_finalized++;
        }
        else
        {
          c_idx[i] = c_idx[i]-1;
          c_final[i] = false;
        }
      }
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
      }
      c_nverts = 8;
      return SV_HEXAHEDRON;
    }
    else if ((line[0] == 'x') && (line[1] == ' '))
    {
      sscanf(&(line[1]), "%d", &(final_idx));
      if (final_idx < 0)
      {
        final_idx = v_count+final_idx;
      }
      else
      {
        final_idx = final_idx - 1;
      }
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
      }
      return SV_FINALIZED;
    }
    else if (line[0] == '#')
    {
      // comments in the body are silently ignored
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
      }
    }
    else
    {
      int j = (int)strlen(line)-1;
      if (j)
      {
        for (int i = 0; i < j; i++)
        {
          if (!isprint(line[i]))
          {
            fprintf(stderr,"WARNING: input file seems to be binary\n");
            break;
          }
        }
        line[(10<j?10:j)] = '\0';
        if (skipped_lines < 5)
        {
          fprintf(stderr,"WARNING: skipping line '%s...'.\n",line);
        }
        else if (skipped_lines == 5)
        {
          fprintf(stderr,"WARNING: skipping more than 5 lines. being quiet.\n");
        }
        skipped_lines++;
      }
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
      }
    }
  }

  if (nverts == -1)
  {
    nverts = v_count;
  }
  else
  {
    if (v_count != nverts)
    {
      fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
    }
  }
  if (ncells == -1)
  {
    ncells = c_count;
  }
  else
  {
    if (c_count != ncells)
    {
      fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
    }
  }
  return SV_EOF;
}

SVevent SVreader_sva::read_event()
{
  if (have_finalized)
  {
    final_idx = finalized_vertices[next_finalized];
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  else
  {
    return read_element();
  }
}

SVreader_sva::SVreader_sva()
{
  // init of SVreader_sva
  file = 0;
  line = 0;
  have_finalized = 0; next_finalized = 0;
  finalized_vertices = 0;
  format = -1;
}

SVreader_sva::~SVreader_sva()
{
  // clean-up for SVreader_sva
}
