/*
===============================================================================

  FILE:  SVreader_vtk.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr
    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2009 martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_vtk.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "vecnfv.h"
#include "vecniv.h"

#define MAX_LINE_SIZE 1024

static void findVTKCellDataPos ( FILE* file, long &where, I32 &c_pnum_f )
{
  fprintf(stderr, "begin findVTKCellDataPos()\n");
  where=-1;
  c_pnum_f = 0;
  //first store the current position
  long curPos = ftell ( file );
  //look for CELL_DATA
  char line[MAX_LINE_SIZE];
  char propName[MAX_LINE_SIZE];
  char dataType[MAX_LINE_SIZE];
  int dummy;
  while ( fgets ( line, MAX_LINE_SIZE, file ) != 0 )
  {
    if ( strncmp ( line,"CELL_DATA ", 10 ) == 0 )
    {
      fprintf(stderr, "found CELL_DATA\n");
      if ( fgets ( line, MAX_LINE_SIZE, file ) != 0 )
      {
        if ( strncmp ( line,"FIELD ", 6 ) == 0 )
        {
          fprintf(stderr, "found FIELD\n");
          if ( fgets ( line, MAX_LINE_SIZE, file ) != 0 )
          {
            if ( sscanf ( line, "%s %d %d %s", propName, &c_pnum_f, &dummy, dataType ) == 4 )
            {
              fprintf(stderr, "Found Cell Property %s (%i components, type %s)\n", propName, c_pnum_f, dataType);
              where = ftell ( file );
              break;
            }
          }
        }
      }
    }
  }
  //set back the current position
  fseek ( file, curPos, SEEK_SET );
}

bool SVreader_vtk::open ( FILE* file )
{
  if ( file == 0 )
  {
    return false;
  }
  this->file = file;

  type = SV_MIXED; // element type not known in advance

  skipped_lines = 0;
  post_order = false;

  line = ( char* ) malloc ( sizeof ( char ) *MAX_LINE_SIZE );

  while ( true )
  {
    if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
    {
      free ( line );
      line = 0;
      return false;
    }

    int i = 0;
    while ( line[i] == ' ' && i < MAX_LINE_SIZE ) i++;

    if ( strncmp ( & ( line[i] ),"POINTS ",7 ) == 0 )
    {
      if ( sscanf ( & ( line[i+7] ), "%d", &nverts ) == 1 )
      {
        break;
      }
      else
      {
        skipped_lines++;
      }
    }
  }

  // read next line
  if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
  {
    free ( line );
    line = 0;
    return false;
  }
  curLinePos = line;

  v_count = 0;
  c_count = 0;

  v_pnum_f = 3;
  v_prop_f = new float[v_pnum_f];

  //see if we have cell data:
  //OK, we found the CELL block, now try to see if there is any cell data.
  findVTKCellDataPos ( file, cellDataCurFilePos, c_pnum_f );
  if(c_pnum_f)
  {
    c_prop_f = new float[c_pnum_f];
  }

  c_idx = new I32[8]; // TET and HEX meshes are supported
  c_final = new bool[8];

  c_final[0] = false;
  c_final[1] = false;
  c_final[2] = false;
  c_final[3] = false;
  c_final[4] = false;
  c_final[5] = false;
  c_final[6] = false;
  c_final[7] = false;

  return true;
}

void SVreader_vtk::close ( bool close_file )
{
  // close of SVreader_vtk
  if ( skipped_lines ) fprintf ( stderr,"WARNING: skipped %d lines.\n",skipped_lines );
  skipped_lines = 0;
  if ( close_file && file && file != stdin ) fclose ( file );
  file = 0;
  if ( line )
  {
    free ( line );
    line = 0;
  }
  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if ( v_prop_f ) delete [] v_prop_f; v_prop_f = 0;
  if ( c_idx ) delete [] c_idx; c_idx = 0;
  if ( c_final ) delete [] c_final; c_final = 0;
}

SVevent SVreader_vtk::read_element()
{
  if ( line == 0 )
  {
    return SV_EOF;
  }

  if ( v_count < nverts )
  {
    v_idx = v_count;
    int nBytesRead;
    while ( sscanf ( curLinePos, "%f %f %f %n", & ( v_prop_f[0] ), & ( v_prop_f[1] ), & ( v_prop_f[2] ), &nBytesRead ) != 3 )
    {
      if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
      {
        fprintf ( stderr,"WARNING: early end-of-file after %d of %d vertices\n", v_count, nverts );
        free ( line );
        line = 0;
        return SV_EOF;
      }
      curLinePos = line;
      skipped_lines++;
    }
    v_count++;
    curLinePos += nBytesRead;
    return SV_VERTEX;
  }
  else if ( c_count == 0 )
  {
    // if we just start reading the cell block we need to skip a few lines
    while ( true )
    {
      if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
      {
        if ( v_count < nverts || c_count < ncells )
        {
          fprintf ( stderr,"WARNING: early end-of-file after %d of %d cells\n", c_count, ncells );
        }
        free ( line );
        line = 0;
        return SV_EOF;
      }
      int i = 0;
      while ( line[i] == ' ' && i < MAX_LINE_SIZE ) i++;

      if ( strncmp ( & ( line[i] ),"CELLS ",6 ) == 0 )
      {
        if ( sscanf ( & ( line[i+6] ), "%d", &ncells ) == 1 )
        {
          break;
        }
        else
        {
          skipped_lines++;
        }
      }
    }
    // read next line
    if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
    {
      if ( c_count < ncells ) fprintf ( stderr,"WARNING: early end-of-file after %d of %d cells\n", c_count, ncells );
      free ( line );
      line = 0;
      return SV_EOF;
    }
  }
  if ( c_count < ncells )
  {
    while ( true )
    {
      if ( sscanf ( & ( line[0] ), "%d", & ( c_nverts ) ) == 1 )
      {
        if ( c_nverts == 4 )
        {
          if ( sscanf ( & ( line[0] ), "%d %d %d %d %d", & ( c_nverts ), & ( c_idx[0] ), & ( c_idx[1] ), & ( c_idx[2] ), & ( c_idx[3] ) ) == 5 )
          {
            c_count++;
            break;
          }
          else
          {
            skipped_lines++;
          }
        }
        else if ( c_nverts == 8 )
        {
          if ( sscanf ( & ( line[0] ), "%d %d %d %d %d %d %d %d %d", & ( c_nverts ), & ( c_idx[0] ), & ( c_idx[1] ), & ( c_idx[2] ), & ( c_idx[3] ), & ( c_idx[4] ), & ( c_idx[5] ), & ( c_idx[6] ), & ( c_idx[7] ) ) == 9 )
            //if (sscanf(&(line[0]), "%d %d %d %d %d %d %d %d %d", &(c_nverts), &(c_idx[0]), &(c_idx[1]), &(c_idx[3]), &(c_idx[2]), &(c_idx[4]), &(c_idx[5]), &(c_idx[7]), &(c_idx[6])) == 9)
          {
            c_count++;
            break;
          }
          else
          {
            skipped_lines++;
          }
        }
        else
        {
          skipped_lines++;
        }
      }
      else
      {
        skipped_lines++;
      }
      if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
      {
        if ( c_count < ncells )
        {
          fprintf ( stderr,"WARNING: early end-of-file after %d of %d cells\n", c_count, ncells );
        }
        free ( line );
        line = 0;
        return SV_EOF;
      }
    }
    //now get the cell data
    if ( c_pnum_f )
    {
      //save file position
      long fpos = ftell ( file );
      //go to cell data position:
      fseek ( file, cellDataCurFilePos, SEEK_SET );
      //read cell data
      int i;
      for( i=0; i<c_pnum_f; ++i)
      {
        if ( fscanf ( file, "%f", & ( c_prop_f[i] )) != 1 )
        {
          skipped_lines++;
          break;
        }
      }
      //save cell data position
      cellDataCurFilePos = ftell ( file );
      //set back file position
      fseek ( file, fpos, SEEK_SET );
    }
    //and get the next cell line
    if ( fgets ( line, sizeof ( char ) * MAX_LINE_SIZE, file ) == 0 )
    {
      if ( c_count < ncells )
      {
        fprintf ( stderr,"WARNING: early end-of-file after %d of %d cells\n", c_count, ncells );
      }
      free ( line );
      line = 0;
    }
    if ( c_nverts == 4 )
    {
      return SV_TETRAHEDRON;
    }
    else
    {
      return SV_HEXAHEDRON;
    }
  }
  free ( line );
  line = 0;
  return SV_EOF;
}

SVevent SVreader_vtk::read_event()
{
  return read_element();
}

SVreader_vtk::SVreader_vtk()
{
  // init of SVreader_vtk
  file = 0;
  line = 0;
}

SVreader_vtk::~SVreader_vtk()
{
  // clean-up for SVreader_vtk
}
