/*
===============================================================================

  FILE: svreadwithboundarymarkers.cpp
  
  CONTENTS:
  
    see corresponding header file

  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2007 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreadwithboundarymarkers.h"

#include <stdlib.h>
#include <stdio.h>

#include "vecnfv.h"
#include "vecniv.h"

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	#include <deque>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
	#include <deque.h>
#endif

#include <assert.h>

#define COLLECT_STATISTICS

struct SVcell;

typedef struct SVvertex
{
  SVvertex* buffer_next;  // used for efficient memory management
  float* prop_f;
  int index;
  bool boundary;
  int list_alloc;
  int list_size;
  SVcell** list;
} SVvertex;

typedef struct SVcell
{
  SVcell* buffer_next;  // used for efficient memory management
  float* prop_f;
  SVvertex* vertices[4];
  int ready;
} SVcell;

typedef hash_map<int, SVvertex*> my_hash;
typedef deque<SVcell*> my_queue;

static my_hash* vertex_hash;
static my_queue* cell_queue;

// efficient memory allocation

static SVvertex* vertex_buffer_next;
static int vertex_buffer_alloc;
static int vertex_buffer_increment;
static int vertex_buffer_initial_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];

#ifdef COLLECT_STATISTICS
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
#endif

static void initVertexBuffer(int alloc)
{
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer_increment = alloc;
  vertex_buffer_initial_increment = alloc;
  vertex_buffer_number_chunks = 0;
#ifdef COLLECT_STATISTICS
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
#endif
}

static SVvertex* allocVertex()
{
  if (vertex_buffer_next == 0)
  {
    vertex_buffer_next = (SVvertex*)malloc(sizeof(SVvertex)*vertex_buffer_increment);
    if (vertex_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreadWithBoundaryMarkers): malloc for vertex buffer with increment %d failed\n", vertex_buffer_increment);
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;
    }
    vertex_buffer_alloc += vertex_buffer_increment;
    for (int i = 0; i < vertex_buffer_increment; i++)
    {
      vertex_buffer_next[i].buffer_next = &(vertex_buffer_next[i+1]);
      vertex_buffer_next[i].prop_f = 0;
      vertex_buffer_next[i].list = 0;
    }
    vertex_buffer_next[vertex_buffer_increment-1].buffer_next = 0;
    vertex_buffer_increment *= 2;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  // initialize data fields
  vertex->index = -1;
  vertex->boundary = false;
  vertex->list_size = 0;

  if (vertex->list == 0)
  {
    vertex->list_alloc = 16;
    vertex->list = (SVcell**)malloc(sizeof(SVcell*)*16);
    if (vertex->list == 0)
    {
      fprintf(stderr,"ERROR (SVreadWithBoundaryMarkers): malloc for vertex->list with %d elements failed\n", 16);
      exit(1);
    }
  }

#ifdef COLLECT_STATISTICS
  // keep track of maxsize
  vertex_buffer_size++; if (vertex_buffer_size > vertex_buffer_maxsize) vertex_buffer_maxsize = vertex_buffer_size;
#endif

  return vertex;
}

static void deallocVertex(SVvertex* vertex)
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef COLLECT_STATISTICS
  vertex_buffer_size--;
#endif
}

static void destroyVertexBuffer()
{
  for (int i = vertex_buffer_number_chunks-1; i >= 0; i--)
  {
    vertex_buffer_increment /= 2;
    for (int j = 0; j < vertex_buffer_increment; j++)
    {
      if (vertex_buffer_chunks[i][j].prop_f) free (vertex_buffer_chunks[i][j].prop_f);
      if (vertex_buffer_chunks[i][j].list) free (vertex_buffer_chunks[i][j].list);
    }
    free(vertex_buffer_chunks[i]);
  }
  vertex_buffer_number_chunks = 0;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
#ifdef COLLECT_STATISTICS
  fprintf(stderr, "vertex_buffer_size %d vertex_buffer_maxsize %d\n", vertex_buffer_size, vertex_buffer_maxsize);
#endif
}

static SVcell* cell_buffer_next;
static int cell_buffer_alloc;
static int cell_buffer_increment;
static int cell_buffer_initial_increment;
static int cell_buffer_number_chunks;
static SVcell* cell_buffer_chunks[16];

#ifdef COLLECT_STATISTICS
static int cell_buffer_size;
static int cell_buffer_maxsize;
#endif

static void initCellBuffer(int alloc)
{
  cell_buffer_next = 0;
  cell_buffer_alloc = 0;
  cell_buffer_increment = alloc;
  cell_buffer_initial_increment = alloc;
  cell_buffer_number_chunks = 0;
#ifdef COLLECT_STATISTICS
  cell_buffer_size = 0;
  cell_buffer_maxsize = 0;
#endif
}

static SVcell* allocCell()
{
  if (cell_buffer_next == 0)
  {
    cell_buffer_next = (SVcell*)malloc(sizeof(SVcell)*cell_buffer_increment);
    if (cell_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreadWithBoundaryMarkers): malloc for cell buffer with increment %d failed\n", cell_buffer_increment);
      exit(1);
    }
    else
    {
      cell_buffer_chunks[cell_buffer_number_chunks] = cell_buffer_next;
      cell_buffer_number_chunks++;
    }
    for (int i = 0; i < cell_buffer_increment; i++)
    {
      cell_buffer_next[i].buffer_next = &(cell_buffer_next[i+1]);
      cell_buffer_next[i].prop_f = 0;
    }
    cell_buffer_next[cell_buffer_increment-1].buffer_next = 0;
    cell_buffer_increment *= 2;
  }

  // get pointer to next available cell
  SVcell* cell = cell_buffer_next;
  cell_buffer_next = cell->buffer_next;
 
  // initialize data fields
  cell->ready = 0;

#ifdef COLLECT_STATISTICS
  // keep track of maxsize
  cell_buffer_size++; if (cell_buffer_size > cell_buffer_maxsize) cell_buffer_maxsize = cell_buffer_size;
#endif

  return cell;
}

static void deallocCell(SVcell* cell)
{
  cell->buffer_next = cell_buffer_next;
  cell_buffer_next = cell;
#ifdef COLLECT_STATISTICS
  cell_buffer_size--;
#endif
}

static void destroyCellBuffer()
{
  for (int i = 0; i < cell_buffer_number_chunks; i++)
  {
    free(cell_buffer_chunks[i]);
  }
  cell_buffer_number_chunks = 0;
  cell_buffer_next = 0;
  cell_buffer_alloc = 0;
#ifdef COLLECT_STATISTICS
  fprintf(stderr, "cell_buffer_size %d cell_buffer_maxsize %d\n", cell_buffer_size, cell_buffer_maxsize);
#endif
}

bool SVreadWithBoundaryMarkers::open(SVreader* svreader)
{
  if (svreader == 0)
  {
    return false;
  }
  this->svreader = svreader;

  nverts = svreader->nverts;
  ncells = svreader->ncells;

  v_count = 0;
  c_count = 0;

  v_pnum_f = svreader->v_pnum_f;
  v_pmin_f = svreader->v_pmin_f;
  v_pmax_f = svreader->v_pmax_f;

  c_pnum_f = svreader->c_pnum_f;
  c_pmin_f = svreader->c_pmin_f;
  c_pmax_f = svreader->c_pmax_f;

  c_idx = new int[4];   // currently only TET meshes are supported
  c_final = new bool[4];
  c_v_prop_f = new float*[4];
  c_boundary = new bool[4];

  type = SV_TET;
  c_nverts = 4;
  post_order = false;

  vertex_hash = new my_hash;
  cell_queue = new my_queue;

  initVertexBuffer(1024);
  initCellBuffer(2048);

  num_non_pre_order = 0;

  return true;
}

void SVreadWithBoundaryMarkers::close(bool close_file)
{
  svreader->close(close_file);

  v_count = -1;
  c_count = -1;

  delete [] c_idx;
  delete [] c_final;
  delete [] c_v_prop_f;
  delete [] c_boundary;

  delete vertex_hash;
  delete cell_queue;

  destroyVertexBuffer();
  destroyCellBuffer();
}

typedef struct SVsort
{
  SVvertex* one;
  SVvertex* two;
} SVsort;

static int sort_array_size = 0;
static SVsort* sort_array = 0;

static void quicksort_for_border_markers(SVsort* a, int i, int j)
{
  int in_i = i;
  int in_j = j;
  SVsort w;
  SVsort key = a[(i+j)/2];
  do
  {
    while ( a[i].one < key.one || (a[i].one == key.one && a[i].two < key.two) ) i++;
    while ( a[j].one > key.one || (a[j].one == key.one && a[j].two > key.two) ) j--;
    if (i<j)
    {
      w = a[i];
      a[i] = a[j];
      a[j] = w;
    }
  } while (++i<=--j);
  if (i == j+3)
  {
    i--;
    j++;
  }
  if (j>in_i) quicksort_for_border_markers(a, in_i, j);
  if (i<in_j) quicksort_for_border_markers(a, i, in_j);
}

static void finalize_vertex(SVvertex* vertex)
{
  int i,j,k;
  SVcell* cell;
  // only vertices that have not yet been marked as boundary need checking
  if (vertex->boundary == false)
  {
    // vertices with an odd number of tetrahedra must be on the boundary
    if (vertex->list_size & 1)
    {
      vertex->boundary = true;
    }
    else
    {
      if (sort_array_size <= 3*vertex->list_size)
      {
        sort_array_size = 3*(vertex->list_size + 10);
        if (sort_array) free (sort_array);
        sort_array = (SVsort*)malloc(sort_array_size*sizeof(SVsort));
      }
      k = 0;
      for (i = 0; i < vertex->list_size; i++)
      {
        cell = vertex->list[i];
        j = 0;
        while (cell->vertices[j] != vertex) j++;
        if (cell->vertices[(j+1)%4] < cell->vertices[(j+2)%4])
        {
          sort_array[k].one = cell->vertices[(j+1)%4]; sort_array[k].two = cell->vertices[(j+2)%4];
        }
        else 
        {
          sort_array[k].one = cell->vertices[(j+2)%4]; sort_array[k].two = cell->vertices[(j+1)%4];
        }
        k++;
        if (cell->vertices[(j+1)%4] < cell->vertices[(j+3)%4])
        {
          sort_array[k].one = cell->vertices[(j+1)%4]; sort_array[k].two = cell->vertices[(j+3)%4];
        }
        else 
        {
          sort_array[k].one = cell->vertices[(j+3)%4]; sort_array[k].two = cell->vertices[(j+1)%4];
        }
        k++;
        if (cell->vertices[(j+2)%4] < cell->vertices[(j+3)%4])
        {
          sort_array[k].one = cell->vertices[(j+2)%4]; sort_array[k].two = cell->vertices[(j+3)%4];
        }
        else 
        {
          sort_array[k].one = cell->vertices[(j+3)%4]; sort_array[k].two = cell->vertices[(j+2)%4];
        }
        k++;
      }
      // set the one past the last entry to zero to simplify the check below
      sort_array[k].one = 0;
      sort_array[k].two = 0;
      // sort the array
      quicksort_for_border_markers(sort_array, 0, k-1);
      // check for boundary faces
      for (i = 0; i < k; i++)
      {
        if (sort_array[i].one != sort_array[i+1].one || sort_array[i].two != sort_array[i+1].two)
        {
          vertex->boundary = true;
          sort_array[i].one->boundary = true;
          sort_array[i].two->boundary = true;      
        }
        else
        {
          i++;
        }
      }
    }
  }
  // check whether this makes any tetrahedra ready for output
  for (i = 0; i < vertex->list_size; i++)
  {
    cell = vertex->list[i];
    if (cell->ready == 3)
    {
      cell_queue->push_back(cell);
    }
    else
    {
      cell->ready++;
    }
  }
}

int SVreadWithBoundaryMarkers::read_input()
{
  int i;
  SVvertex* vertex=NULL;
  SVcell* cell;
  my_hash::iterator hash_element;

  SVevent event = svreader->read_element();
    
  if (event == SV_TETRAHEDRON)
  {
    assert(svreader->c_nverts == 4);
    // create cell
    cell = allocCell();
    // copy cell's float properties
    if (c_pnum_f)
    {
      if (cell->prop_f == 0) cell->prop_f = new float[c_pnum_f];
      VecCopyNfv(cell->prop_f, svreader->c_prop_f, c_pnum_f);
    }
    // process the cell's vertices
    for (i = 0; i < 4; i++)
    {
      // look for vertex in the hash
      hash_element = vertex_hash->find(svreader->c_idx[i]);
      // if vertex not in hash we have something other than a pre-order mesh  
      if (hash_element == vertex_hash->end())
      {
        // create a vertex and insert into hash
        vertex = allocVertex();
        // insert vertex into hash
        vertex_hash->insert(my_hash::value_type(svreader->c_idx[i], vertex));
        // increment the counter for the number vertices that are not in pre-order
        num_non_pre_order++;
      }
      else
      {
        // use vertex found in hash
        vertex = (*hash_element).second;
        // maybe we should remove it
        if (svreader->c_final[i]) vertex_hash->erase(hash_element);
      }
      // add vertex to cell
      cell->vertices[i] = vertex;
      // add tetrahedron to vertex
      if (vertex->list_size == vertex->list_alloc)
      {
        vertex->list_alloc += 4;
        vertex->list = (SVcell**)realloc(vertex->list, sizeof(SVcell*)*vertex->list_alloc);
        if (!vertex->list)
        {
          fprintf(stderr, "ERROR (SVreadWithBoundaryMarkers): realloc for %d elements of vertex->list failed\n", vertex->list_alloc);
          return 0;
        }
      }
      vertex->list[vertex->list_size] = cell;
      vertex->list_size++;
    }
    //  // does this cell finalize any vertices
    for (i = 0; i < 4; i++)
    {
      // does this cell finalize this vertex
      if (svreader->c_final[i])
      {
        // finalize the vertex
        finalize_vertex(cell->vertices[i]);
      }
    }
  }
  else if (event == SV_VERTEX)
  {
    if (num_non_pre_order)
    {
      // look for vertex in the hash
      hash_element = vertex_hash->find(svreader->v_idx);
      if (hash_element == vertex_hash->end())
      {
        // create a vertex and insert into hash
        vertex = allocVertex();
        // insert vertex into hash
        vertex_hash->insert(my_hash::value_type(svreader->v_idx, vertex));
      }
      else
      {
        // decrement the counter for the number vertices that are not in pre-order
        num_non_pre_order--;
      }
      // copy vertex's float properties
      if (vertex->prop_f == 0) vertex->prop_f = new float[v_pnum_f];
      VecCopyNfv(vertex->prop_f, svreader->v_prop_f, v_pnum_f);
    }
    else
    {
      // create a vertex and insert into hash
      vertex = allocVertex();
      // insert vertex into hash
      vertex_hash->insert(my_hash::value_type(svreader->v_idx, vertex));
      // copy vertex's float properties
      if (vertex->prop_f == 0) vertex->prop_f = new float[v_pnum_f];
      VecCopyNfv(vertex->prop_f, svreader->v_prop_f, v_pnum_f);
    }
    // for post-order input this also finalizes the vertex
    if (svreader->post_order)
    {
      finalize_vertex(vertex);
    }
  }
  else if (event == SV_HEXAHEDRON)
  {
    assert(svreader->c_nverts == 8);
  }
  else if (event == SV_FINALIZED)
  {
    // look for finalized vertex in the vertex hash
    hash_element = vertex_hash->find(svreader->final_idx);
    // vertices must preceed their finalization
    if (hash_element == vertex_hash->end())
    {
      fprintf(stderr, "ERROR (SVreadWithBoundaryMarkers): explicitely finalized vertex %d not in hash.\n", svreader->final_idx);
      return 0;
    }
    // get vertex found in hash
    vertex = (*hash_element).second;
    // erase its entry from the hash
    vertex_hash->erase(hash_element);
    // and finalize it
    finalize_vertex(vertex); 
    return 1;
  }
  else if (event == SV_EOF)
  {
    // all remaining vertices are implicitely finalized
    if (vertex_hash->size())
    {
      // get some hash element ... 
      hash_element = vertex_hash->begin();
      // get the vertex
      vertex = (*hash_element).second;
      // erase its entry from the hash
      vertex_hash->erase(hash_element);
      // and finalize it
      finalize_vertex(vertex); 
    }
    else
    {
      return 0;
    }
  }
  return 1;
}

int SVreadWithBoundaryMarkers::read_cell()
{
  int i;
  SVvertex* vertex;
  SVcell* cell;

  // read from input until tetrahedra are ready for ouput

  while (!cell_queue->size())
  {
    int ok = read_input();
    if (ok <= 0)
    {
      return ok;
    }
  }

  // get next output tetrahedron

  cell = cell_queue->front();
  cell_queue->pop_front();

  // now we have a tetrahedron

  have_cell = 1;

  // look for new or finalized vertices and fill in indices, finalization, and positions

  for (i = 0; i < 4; i++)
  {
    vertex = cell->vertices[i];
    // is this vertex a new vertex?
    if (vertex->index == -1)
    {
      vertex->index = v_count + have_new;
      new_vertices[have_new++] = i;
    }
    assert(vertex->index >= 0);
    // is this vertex to be finalized?
    if (vertex->list_size == 1)
    {
      c_final[i] = true;
      finalized_vertices[have_finalized++] = i;
      deallocVertex(vertex); // the memory can still be used until the next vertex is allocated.
    }
    else
    {
      vertex->list_size--;
      c_final[i] = false;
    }
    // fill index array
    c_idx[i] = vertex->index;
    // set additional vertex property array
    c_v_prop_f[i] = vertex->prop_f;
    // fill boundary marker array
    c_boundary[i] = vertex->boundary;
  }

  // set cell properties

  c_prop_f = cell->prop_f;

  // deallocate this cell

  deallocCell(cell); // the memory can still be used until the next cell is allocated.

  return 1;
}

SVevent SVreadWithBoundaryMarkers::read_element()
{
  if ((have_new + have_cell) == 0)
  {
    next_new = 0;
    have_finalized = next_finalized = 0;
    
    int ok = read_cell();
    if (ok <= 0)
    {
      return (SVevent)ok;
    }
  }
  if (have_new)
  {
    // reference vertex properties in API interface
    v_idx = c_idx[new_vertices[next_new]];
    v_prop_f = c_v_prop_f[new_vertices[next_new]];
    v_boundary = c_boundary[new_vertices[next_new]];
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else
  {
    // reference cell properties in API interface
    have_cell = 0;
    c_count++;
    return SV_TETRAHEDRON;
  }
}

SVevent SVreadWithBoundaryMarkers::read_event()
{
  if ((have_new + have_cell + have_finalized) == 0)
  {
    return read_element();
  }
  if (have_new)
  {
    // reference vertex properties in API interface
    v_idx = c_idx[new_vertices[next_new]];
    v_prop_f = c_v_prop_f[new_vertices[next_new]];
    v_boundary = c_boundary[new_vertices[next_new]];
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if (have_cell)
  {
    // reference cell properties in API interface
    have_cell = 0;
    c_count++;
    return SV_TETRAHEDRON;
  }
  else if (have_finalized)
  {
    final_idx = c_idx[finalized_vertices[next_finalized]];
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  return SV_ERROR;
}

SVreadWithBoundaryMarkers::SVreadWithBoundaryMarkers()
{
  // init of SVreadWithBoundaryMarkers
  have_new = 0; next_new = 0;
  have_cell = 0;
  have_finalized = 0; next_finalized = 0;

  svreader = 0;
  num_non_pre_order = 0;
}

SVreadWithBoundaryMarkers::~SVreadWithBoundaryMarkers()
{
  // clean-up for SVreadWithBoundaryMarkers
  v_prop_f = 0;
  v_pnum_f = 0;
  v_pmin_f = 0;
  v_pmax_f = 0;
  c_prop_f = 0;
  c_pnum_f = 0;
  c_pmin_f = 0;
  c_pmax_f = 0;
  if (svreader) delete svreader;
}
