/*
===============================================================================

  FILE:  svwriteopener.cpp
  
  CONTENTS:
  
    see corresponding header file
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2007  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svwriteopener.h"

// streaming mesh writers
#include "svwriter_sva.h"
#include "svwriter_svb.h"
#include "svwriter_svc.h"
#include "svwriter_svh.h"
#include "svwriter_svg.h"
#ifdef BUILD_HZIP
#include "svwriter_hzip.h"
#endif

//special writers
#include "svwriter_nil.h"

// non-streaming mesh writers
#include "svwriter_off.h"
#include "svwriter_vtk.h"

// for buffered writing
#include "svwritebuffered.h"
#include "svwritebufferedhex.h"

#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
extern "C" FILE* fopenGzipped(const char* filename, const char* mode);
#endif

void SVwriteOpener::set_file_format(const char* file_format)
{
  if (this->file_format) free(this->file_format);
  this->file_format = strdup(file_format);
}

void SVwriteOpener::set_file_name(const char* file_name)
{
  if (this->file_name) free(this->file_name);
  this->file_name = strdup(file_name);
}

void SVwriteOpener::set_cbm_precision()
{
  this->cbm_precision = true;
}

void SVwriteOpener::set_compression_bits(int compression_bits)
{
  this->compression_bits = compression_bits;
}

void SVwriteOpener::set_high_bits(int high_bits)
{
  this->high_bits = high_bits;
}

void SVwriteOpener::set_celldata_compression_bits(int compression_bits)
{
  this->celldata_compression_bits = compression_bits;
}

void SVwriteOpener::set_celldata_high_bits(int high_bits)
{
  this->celldata_high_bits = high_bits;
}

void SVwriteOpener::set_maximum_delay(int maximum_delay)
{
  this->maximum_delay = maximum_delay;
}

void SVwriteOpener::set_dry_run()
{
  this->dry_run = true;
}

SVwriter* SVwriteOpener::open()
{
  // if the file format is not given explicitely we try to guess it
  if (file_format == 0)
  {
    if (file_name == 0)
    {
      fprintf(stderr, "ERROR: neither a file name nor a file format were specified\n");
      return 0;
    }
    else if (strstr(file_name, ".sva"))
      file_format = strdup("sva");
    else if (strstr(file_name, ".svb"))
      file_format = strdup("svb");
    else if (strstr(file_name, ".svc"))
      file_format = strdup("svc");
    else if (strstr(file_name, ".svh"))
      file_format = strdup("svh");
    else if (strstr(file_name, ".svg"))
      file_format = strdup("svg");
#ifdef BUILD_HZIP
    else if (strstr(file_name, ".hzip"))
      file_format = strdup("hzip");
#endif
    else if (strstr(file_name, ".off"))
      file_format = strdup("off");
    else if (strstr(file_name, ".vtk"))
      file_format = strdup("vtk");
    else if (strstr(file_name, "nil"))
      file_format = strdup("nil");
    else
    {
      fprintf(stderr, "ERROR: could not guess file format from file name '%s'\n", file_name);
      return 0;
    }
  }

  // depending on the file format we open the file in binary or text mode
  bool is_binary = true;
  if (strcmp(file_format, "sva") == 0 || strcmp(file_format, "off") == 0 || strcmp(file_format, "vtk") == 0)
    is_binary = false;

  FILE* file;

  // if we have no file name we write to stdout, otherwise we open the file
  if (file_name == 0)
  {
    file = stdout;
#ifdef _WIN32
    if (is_binary)
      if(_setmode( _fileno( stdout ), _O_BINARY ) == -1 )
      {
        fprintf(stderr, "ERROR: cannot set stdout to binary (untranslated) mode\n");
      }
    else
      if(_setmode( _fileno( stdout ), _O_TEXT ) == -1 )
      {
        fprintf(stderr, "ERROR: cannot set stdout to text (translated) mode\n");
      }
#endif
  }
  else
  {
    if (strstr(file_name, ".gz"))
    {
#ifdef _WIN32
      file = fopenGzipped(file_name, (is_binary ? "wb" : "w"));
#else
      fprintf(stderr,"ERROR: cannot open gzipped file '%s'\n",file_name);
      return 0;
#endif
    }
    else
    {
      file = fopen(file_name, (is_binary ? "wb" : "w"));
    }
    if (file == 0)
    {
      fprintf(stderr,"ERROR: cannot open file '%s'\n",file_name);
      return 0;
    }
  }

  // now we open the respective writer

  if (strcmp(file_format, "sva") == 0)
  {
    SVwriter_sva* svwriter_sva = new SVwriter_sva();
    if (svwriter_sva->open(file)) return svwriter_sva;
    delete svwriter_sva;
  }
  else if (strcmp(file_format, "svb") == 0)
  {
    SVwriter_svb* svwriter_svb = new SVwriter_svb();
    if (svwriter_svb->open(file)) return svwriter_svb;
    delete svwriter_svb;
  }
  else if (strcmp(file_format, "svc") == 0)
  {
    SVwriter_svc* svwriter_svc = new SVwriter_svc();
    if (svwriter_svc->open(dry_run ? 0 : file, compression_bits,cbm_precision,high_bits))
    {
      if (maximum_delay)
      {
        SVwriteBuffered* svwrite_buffered = new SVwriteBuffered();
        if (svwrite_buffered->open(svwriter_svc, maximum_delay)) return svwrite_buffered;
        delete svwrite_buffered;
      }
      else
      {
        return svwriter_svc;
      }
    }
    else
    {
      delete svwriter_svc;
    }
  }
  else if (strcmp(file_format, "svh") == 0)
  {
    SVwriter_svh* svwriter_svh = new SVwriter_svh();
    if (svwriter_svh->open(dry_run ? 0 : file, compression_bits, high_bits, celldata_compression_bits, celldata_high_bits))
    {
      if (maximum_delay)
      {
        SVwriteBufferedHex* svwrite_buffered = new SVwriteBufferedHex();
        if (svwrite_buffered->open(svwriter_svh, maximum_delay)) return svwrite_buffered;
        delete svwrite_buffered;
      }
      else
      {
        return svwriter_svh;
      }
    }
    else
    {
      delete svwriter_svh;
    }
  }
  else if (strcmp(file_format, "svg") == 0)
  {
    SVwriter_svg* svwriter_svg = new SVwriter_svg();
    if (svwriter_svg->open(file))
    {
      return svwriter_svg;
    }
    else
    {
      delete svwriter_svg;
    }
  }
#ifdef BUILD_HZIP
  else if (strcmp(file_format, "hzip") == 0)
  {
    SVwriter_hzip* svwriter_hzip = new SVwriter_hzip();
    if (svwriter_hzip->open(file))
    {
      return svwriter_hzip;
    }
    else
    {
      delete svwriter_hzip;
    }
  }
#endif
  else if (strcmp(file_format, "off") == 0)
  {
    SVwriter_off* svwriter_off = new SVwriter_off();
    if (svwriter_off->open(file)) return svwriter_off;
    delete svwriter_off;
  }
  else if (strcmp(file_format, "vtk") == 0)
  {
    SVwriter_vtk* svwriter_vtk = new SVwriter_vtk();
    if (svwriter_vtk->open(file)) return svwriter_vtk;
    delete svwriter_vtk;
  }
  else if (strcmp(file_format, "nil") == 0)
  {
    SVwriter_nil* svwriter_nil = new SVwriter_nil();
    if (svwriter_nil->open(file)) return svwriter_nil;
    delete svwriter_nil;
  }
  else
  {
    fprintf(stderr,"ERROR: cannot determine the writer for format '%s'\n", file_format);
  }
  return 0;
}

SVwriteOpener::SVwriteOpener()
{
  file_format = 0;
  file_name = 0;
  compression_bits = 16;
  high_bits = 8;
  celldata_compression_bits = 0;
  celldata_high_bits = 8;
  maximum_delay = 0;
  dry_run = false;
  cbm_precision = false;
}

SVwriteOpener::~SVwriteOpener()
{
  if (file_format) free(file_format);
  if (file_name) free(file_name);
}
