/*
===============================================================================

  FILE:  SVwriter_svg.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_svg.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#if !(defined(i386) || defined(_WIN32) || defined(AMD64))
#define BIG_ENDIAN_SYSTEM
#endif

#include <stdlib.h>
#include <string.h>

//#define PRINT_DEBUG_OUTPUT

#define SVG_VERSION 1

typedef enum
{
	SVG_VERTEX = 0,
	SVG_TETRAHEDRON = 1,
	SVG_HEXAHEDRON = 2
} SVElementType;

//#define USE_ASM_BYTESWAP
#ifdef USE_ASM_BYTESWAP
extern inline U32 swap_endianness ( const U32 *x );
#else
inline U32 swap_endianness ( const U32 *x )
{
	U32 xx = ( *x>>24 ) |
	              ( ( *x<<8 ) & 0x00FF0000 ) |
	              ( ( *x>>8 ) & 0x0000FF00 ) |
	              ( *x<<24 );
	return xx;
}
#endif


inline void SVwriter_svg::writeInt32 ( const I32 *i )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "writing I32\t%i\n", *i);
#endif
#ifdef BIG_ENDIAN_SYSTEM
	U32 ii = swap_endianness ( ( U32* ) i );
	fwrite ( &ii, 4, 1, file );
#else
	fwrite ( i, 4, 1, file );
#endif
}

inline void SVwriter_svg::writeFloat ( const float *f )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "writing float\t%g\n", *f);
#endif
#ifdef BIG_ENDIAN_SYSTEM
	U32 ii = swap_endianness ( ( U32* ) f );
	fwrite ( &ii, 4, 1, file );
#else
	fwrite ( f, 4, 1, file );
#endif
}

inline void SVwriter_svg::writeNFloats ( const float *f, int n )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "writing %i floats\n", n);
#endif
#ifdef BIG_ENDIAN_SYSTEM
	const float *fp = f;
	for ( int i=0; i<n; ++i, ++fp )
	{
		U32 ii = swap_endianness ( ( U32* ) fp );
		fwrite ( &ii, 4, 1, file );
	}
#else
	fwrite ( f, 4, n, file );
#endif
}

inline void SVwriter_svg::writeUChar ( const unsigned char *c )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "writing uint8_t\t%i\n", (int)*c);
#endif
	fwrite ( c, 1, 1, file );
}

bool SVwriter_svg::open ( FILE* file )
{
	if ( file == 0 )
	{
		fprintf ( stderr, "ERROR: zero file pointer not supported by SVwriter_svg\n" );
		exit ( 0 );
	}

#ifdef _WIN32
	if ( file == stdout )
	{
		if ( _setmode ( _fileno ( stdout ), _O_BINARY ) == -1 )
		{
			fprintf ( stderr, "ERROR: cannot set stdout to binary (untranslated) mode\n" );
			exit ( 0 );
		}
	}
#endif

	this->file = file;
	// write version
	unsigned char version = SVG_VERSION;
	writeUChar ( &version );

	v_count = 0;
	c_count = 0;

	return true;
}

void SVwriter_svg::close ( bool close_file, bool update )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "close() called\n");
#endif
	// close of SVwriter_svg
	if ( nverts != v_count || ncells != c_count )
	{
		if ( update && file != stdout ) update_header();
	}
	if ( close_file && file && file != stdout ) fclose ( file );
	file = 0;

	// close of SVwriter interface
	if ( nverts != -1 && nverts != v_count )  
		fprintf ( stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count );
	if ( ncells != -1 && ncells != c_count )  
		fprintf ( stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count );
	nverts = v_count;
	ncells = c_count;
	v_count = -1;
	c_count = -1;
}

void SVwriter_svg::write_vertex ( const float* v_prop_f )
{
	if ( v_count + c_count == 0 ) write_header();
	//write element type
	unsigned char elementType = SVG_VERTEX;
	writeUChar(&elementType);
	//write vertex data
	if ( v_pnum_f )
	{
		writeNFloats ( v_prop_f, v_pnum_f );
	}
	v_count++;
}

void SVwriter_svg::write_tetrahedron ( const I32* c_idx, const bool* c_final )
{
	write_tetrahedron ( c_idx, NULL, c_final );
}

void SVwriter_svg::write_tetrahedron ( const I32* c_idx, const float* c_prop_f, const bool* c_final )
{
	if ( v_count + c_count == 0 ) write_header();
	//write element type
	unsigned char elementType = SVG_TETRAHEDRON;
	writeUChar(&elementType);
	//write indices
	for(int i=0; i<4; ++i)
	{
		I32 index = c_final[i] ? c_idx[i]-v_count : c_idx[i]+1;
		writeInt32(&index);
	}	
	//write cell data
	if ( c_pnum_f )
	{
		writeNFloats ( c_prop_f, c_pnum_f );
	}
	c_count++;
}

void SVwriter_svg::write_hexahedron ( const I32* c_idx, const bool* c_final )
{
	write_hexahedron ( c_idx, NULL, c_final );
}

void SVwriter_svg::write_hexahedron ( const I32* c_idx, const float* c_prop_f, const bool* c_final )
{
	if ( v_count + c_count == 0 ) write_header();
	//write element type
	unsigned char elementType = SVG_HEXAHEDRON;
	writeUChar(&elementType);
	//write indices
	for(int i=0; i<8; ++i)
	{
		I32 index = c_final[i] ? c_idx[i]-v_count : c_idx[i]+1;
		writeInt32(&index);
	}	
	//write cell data
	if ( c_pnum_f )
	{
		writeNFloats ( c_prop_f, c_pnum_f );
	}
	c_count++;
}

void SVwriter_svg::write_finalized ( I32 final_idx )
{
	fprintf ( stderr, "ERROR: write_finalized(int final_idx) not supported by SVwriter_svg\n" );
	exit ( 0 );
}

void SVwriter_svg::write_header()
{
	// record where nverts and nfaces are located
	update_seek_pos = ( U32 ) ftell ( file );
	// write nverts
	writeInt32( &nverts );
	writeInt32( &ncells );
	// write how many floating-point attributes (per vertex)
	writeInt32( &v_pnum_f );
	// write how many floating-point attributes (per cell)
	writeInt32( &c_pnum_f );
	// write bounding box of floating-point attributes (per vertex)
	if ( v_pnum_f )
	{
		if ( v_pmin_f && v_pmax_f )
		{
			writeNFloats( v_pmin_f, v_pnum_f );
			writeNFloats( v_pmax_f, v_pnum_f );
		}
		else
		{
			fprintf ( stderr, "ERROR: SVwriter_svg requires bounding box information for vertex properties. Consider using -compute_bb\n" );
			exit(1);
		}
	}
	// write bounding box of floating-point attributes (per cell)
	if ( c_pnum_f )
	{
		if ( c_pmin_f && c_pmax_f )
		{
			writeNFloats( c_pmin_f, c_pnum_f );
			writeNFloats( c_pmax_f, c_pnum_f );
		}
		else
		{
			fprintf ( stderr, "ERROR: SVwriter_svg requires bounding box information for cell properties. Consider using -compute_bb\n" );
			exit(1);
		}
	}
	//write dummy information
	I32 dummy = 0;
	writeInt32(&dummy);
	writeInt32(&dummy);
}

void SVwriter_svg::update_header()
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "update_header() called\n");
#endif
	if ( file == stdout )
	{
		fprintf ( stderr,"WARNING: cannot update header when writing to stdout\n" );
		return;
	}
	if ( fseek ( file, update_seek_pos, SEEK_SET ) != 0 )
	{
		fprintf ( stderr,"WARNING: cannot update header because fseek failed\n" );
		return;
	}
	writeInt32(&v_count);
	writeInt32(&c_count);
	if ( fseek ( file, 0L, SEEK_END ) != 0 )
	{
		fprintf ( stderr,"WARNING: header was updated but final fseek failed\n" );
	}
}

SVwriter_svg::SVwriter_svg()
{
	// init of SVwriter_svg
	file = 0;
}

SVwriter_svg::~SVwriter_svg()
{
	// clean-up for SVwriter_svg
}
