/*
===============================================================================

  FILE:  SVwriter_vtk.cpp
  
  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2005-07 martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_vtk.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "vecnfv.h"
#include "vecniv.h"

bool SVwriter_vtk::open(FILE* file)
{
  if (file == 0)
  {
    return false;
  }

#ifdef _WIN32
  if (file == stdout)
  {
    if(_setmode( _fileno( stdout ), _O_TEXT ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdout to text (translated) mode\n");
    }
  }
#endif

  this->file = file;
  v_count = 0;
  c_count = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer = 0;
  cell_buffer_alloc = 0;
  cell_buffer = 0;
  cellData_buffer = 0;
  return true;
}

void SVwriter_vtk::write_vertex(const float* v_prop_f)
{
  if (v_count == vertex_buffer_alloc)
  {
    if (vertex_buffer_alloc == 0)
    {
      if (nverts != -1)
      {
        vertex_buffer = (float*)realloc(vertex_buffer, sizeof(float)*v_pnum_f*nverts);
        vertex_buffer_alloc = nverts;
      }
      else
      {
        vertex_buffer = (float*)realloc(vertex_buffer, sizeof(float)*v_pnum_f*1024);
        vertex_buffer_alloc = 1024;
      }
    }
    else
    {
      vertex_buffer = (float*)realloc(vertex_buffer, sizeof(float)*v_pnum_f*vertex_buffer_alloc*2);
      vertex_buffer_alloc = vertex_buffer_alloc * 2;
    }
  }
  VecCopyNfv(&(vertex_buffer[v_count*v_pnum_f]),v_prop_f,v_pnum_f);

  v_count++;
}

void SVwriter_vtk::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  if (c_count == cell_buffer_alloc)
  {
    if (cell_buffer_alloc == 0)
    {
      if (ncells != -1)
      {
        cell_buffer = (I32*)malloc(sizeof(I32)*8*ncells);
        cell_buffer_alloc = ncells;
      }
      else
      {
        cell_buffer = (I32*)malloc(sizeof(I32)*8*2048);
        cell_buffer_alloc = 2048;
      }
    }
    else
    {
      cell_buffer = (I32*)realloc(cell_buffer, sizeof(I32)*8*cell_buffer_alloc*2);
      cell_buffer_alloc = cell_buffer_alloc * 2;
    }
  }
  VecCopyNiv(&(cell_buffer[c_count*8]),c_idx,4);
  cell_buffer[c_count*8+7] = -99;

  c_count++;
}

void SVwriter_vtk::write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  write_tetrahedron(c_idx, c_final);
}

void SVwriter_vtk::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  write_hexahedron(c_idx, NULL, c_final);
}

void SVwriter_vtk::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  if (c_count == cell_buffer_alloc)
  {
    if (cell_buffer_alloc == 0)
    {
      if (ncells != -1)
      {
        cell_buffer = (I32*)malloc(sizeof(I32)*8*ncells);
        if(c_pnum_f)
        {
          cellData_buffer = (float*)malloc(sizeof(float)*c_pnum_f*ncells);
        }
        cell_buffer_alloc = ncells;
      }
      else
      {
        cell_buffer = (I32*)malloc(sizeof(I32)*8*2048);
        if(c_pnum_f)
        {
          cellData_buffer = (float*)malloc(sizeof(float)*c_pnum_f*2048);
        }
        cell_buffer_alloc = 2048;
      }
    }
    else
    {
      cell_buffer = (I32*)realloc(cell_buffer, sizeof(I32)*8*cell_buffer_alloc*2);
      if(c_pnum_f)
      {
        cellData_buffer = (float*)realloc(cellData_buffer, sizeof(float)*c_pnum_f*cell_buffer_alloc*2);
      }
      cell_buffer_alloc = cell_buffer_alloc * 2;
    }
  }
  VecCopyNiv(&(cell_buffer[c_count*8]), c_idx, 8);
  if(c_pnum_f)
  {
  	VecCopyNfv(&(cellData_buffer[c_count*c_pnum_f]), c_prop_f, c_pnum_f);
  }

  c_count++;
}

void SVwriter_vtk::write_finalized(I32 final_idx)
{
}

void SVwriter_vtk::close(bool close_file, bool update_header)
{
  // close SVwriter_vtk
  int i;
  // write header
  fprintf(file, "# vtk DataFile Version 2.0\012");
  fprintf(file, "generated by Martin's streaming volume mesh library\012");
  fprintf(file, "ASCII\012");
  fprintf(file, "DATASET UNSTRUCTURED_GRID\012");
  // write vertices
  fprintf(file, "POINTS %d float\012",v_count);
  for(i = 0; i < v_count; i++)
  {
    fprintf(file, "%.8g %.8g %.8g\012",vertex_buffer[v_pnum_f*i+0],vertex_buffer[v_pnum_f*i+1],vertex_buffer[v_pnum_f*i+2]);
  }
  // write cells
  //first compute the number of total integer values needed to store CELLS:
  int connSize=0;
  for(i = 0; i < c_count; i++)
  {
    if (cell_buffer[8*i+7] == -99)
      connSize+=5;
    else
      connSize+=9;
  }
  fprintf(file, "CELLS %d %d\012",c_count, connSize);
  for(i = 0; i < c_count; i++)
  {
    if (cell_buffer[8*i+7] == -99)
    {
      // write a tet
      fprintf(file, "4 %d %d %d %d\012",cell_buffer[8*i+0],cell_buffer[8*i+1],cell_buffer[8*i+2],cell_buffer[8*i+3]);
    }
    else
    {
      // write a hex
      fprintf(file, "8 %d %d %d %d %d %d %d %d\012",cell_buffer[8*i+0],cell_buffer[8*i+1],cell_buffer[8*i+2],cell_buffer[8*i+3],cell_buffer[8*i+4],cell_buffer[8*i+5],cell_buffer[8*i+6],cell_buffer[8*i+7]);
    }
  }
  // write cell types
  fprintf(file, "CELL_TYPES %d\012",c_count);
  for(i = 0; i < c_count; i++)
  {
    if (cell_buffer[8*i+7] == -99)
    {
      // write a tet type
      fprintf(file, "10\012");
    }
    else
    {
      // write a hex type
      fprintf(file, "12\012");
    }
  }

  // write data
  if(c_pnum_f)
  {
    fprintf(file, "CELL_DATA %d\012",c_count);
    fprintf(file, "FIELD FieldData %d\012",c_pnum_f);
    for(int array=0; array<c_pnum_f; ++array)
    {
      fprintf(file, "CellProperty%d 1 %d float\012", array, c_count);
      int cdIndex = array;
      for(i = 0; i < c_count; i++, cdIndex+=c_pnum_f)
          fprintf(file, "%f\012", cellData_buffer[cdIndex]);
    }
  }

  vertex_buffer_alloc = 0;
  if (vertex_buffer) free(vertex_buffer);
  vertex_buffer = 0;

  cell_buffer_alloc = 0;
  if (cell_buffer) free(cell_buffer);
  cell_buffer = 0;
  if (cellData_buffer) free(cellData_buffer);
  cellData_buffer = 0;

  if (close_file && file && file != stdout) fclose(file);
  file = 0;

  // close of SVwriter interface
  if (nverts != -1) if (nverts != v_count)  fprintf(stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count)  fprintf(stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;
}

SVwriter_vtk::SVwriter_vtk()
{
  // init of SVwriter_vtk interface
  file = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer = 0;
  cell_buffer_alloc = 0;
  cell_buffer = 0;
}

SVwriter_vtk::~SVwriter_vtk()
{
  // clean-up for SVwriter_vtk
}
