/*
===============================================================================

  FILE:  sv_diagram.cpp

  CONTENTS:

    This little tool visualizes the layout of a (streaming) volume mesh.

  PROGRAMMERS:

    martin isenburg@cs.unc.edu
    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    4 Novermber 2009 -- (clement) Added support for Hex meshes
    15 October 2007 -- (martin) using svreadopener for file IO
    26 June 2005 -- (martin) quickly hacked from sm_diagram.cpp for the MPI talk

===============================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <GL/glut.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
#include <ext/hash_map>
namespace std { using namespace __gnu_cxx; }
using namespace std;
#else
#include <hash_map.h>
#endif

#include "vecniv.h"

#include "svreadopener.h"

// MOUSE INTERFACE

static int LeftButtonDown=0;
static int MiddleButtonDown=0;
static int RightButtonDown=0;
static int OldX,OldY;//,,NewY;
static float DistX=0;
static float DistY=0;
static float DistZ=1.2f;

// VISUALIZATION SETTINGS

//static float boundingBoxScale=1.0f;
//static float boundingBoxTranslateX = 0.0f;
//static float boundingBoxTranslateY = 0.0f;
//static float boundingBoxTranslateZ = 0.0f;

// GLOBAL CONTROL VARIABLES

static int WindowW=1024, WindowH=768;
//static int WindowW=800, WindowH=600;
//static int WindowW=1280, WindowH=960;
static int InteractionMode=1;
static int AnimationOn=0;
static int WorkingOn=0;
static int PrintToFileOn=0;

// COLORS

static float colours_diffuse[10][4];

// DATA STORAGE FOR STREAM VISUALIZER OUTPUT

static SVreadOpener* svreadopener = 0;
static SVreader* svreader = 0;

static int nverts;
static int ncells;

static int EXACTLY_N_CELLS = 500000;
static int EVERY_NTH_CELL = -1;
static int TOTAL_CELLS;
static int NEXT_CELL;

static int EXACTLY_N_STEPS = 25;
static int EVERY_NTH_STEP = -1;
static int TOTAL_STEPS;
static int NEXT_STEP;


static float aspect_ratio = 0.2f;

static float POINTSIZE = 1.0f;
static float LINEWIDTH_MAXSPANS = 3.0f;
static float LINEWIDTH_BOX = 2.0f;

static int SHOW_VERTEX_SPAN=1;
static int SHOW_CELL_SPAN=1;
static int SHOW_REFERENCES=1;

static int SHOW_MAX_SPANS=0;

static unsigned char* Framebuffer = 0;
static const char* PrintFileName = "frame";
static int Time=0;

// efficient memory allocation

typedef struct Span
{
	int start;
	int end;
} Span;

// DATA STORAGE FOR STREAM VISUALIZER OUTPUT

static int illustrated_cells_num = 0;
static int illustrated_cells_alloc = 0;
static int* illustrated_cells = 0;
//illustrated_cells[10*i] is the type of i-th cell
//illustrated_cells[10*i+1] is the index of i-th cell
//illustrated_cells[10*i+k], k>1 are the vertex indices

static void SavePPM ( char *FileName, unsigned char* Colour, int Width, int Height )
{
	FILE *fp = fopen ( FileName, "wb" );
	fprintf ( fp, "P6\n%d %d\n255\n", Width, Height );
	int NumRowPixels = Width*3;
	for ( int i= ( Height-1 ) *Width*3; i>=0; i-= ( Width*3 ) )
	{
		fwrite ( & ( Colour[i] ),1,NumRowPixels,fp );
	}
	fclose ( fp );
}

static void InitColors()
{
	colours_diffuse[0][0] = 0.0f; colours_diffuse[0][1] = 0.0f; colours_diffuse[0][2] = 0.0f; colours_diffuse[0][3] = 1.0f; // black
	colours_diffuse[1][0] = 0.6f; colours_diffuse[1][1] = 0.0f; colours_diffuse[1][2] = 0.0f; colours_diffuse[1][3] = 1.0f; // red
	colours_diffuse[2][0] = 0.0f; colours_diffuse[2][1] = 0.8f; colours_diffuse[2][2] = 0.0f; colours_diffuse[2][3] = 1.0f; // green
	colours_diffuse[3][0] = 0.0f; colours_diffuse[3][1] = 0.0f; colours_diffuse[3][2] = 0.6f; colours_diffuse[3][3] = 1.0f; // blue
	colours_diffuse[4][0] = 0.6f; colours_diffuse[4][1] = 0.6f; colours_diffuse[4][2] = 0.0f; colours_diffuse[4][3] = 1.0f; // yellow
	colours_diffuse[5][0] = 0.6f; colours_diffuse[5][1] = 0.0f; colours_diffuse[5][2] = 0.6f; colours_diffuse[5][3] = 1.0f; // purple
	colours_diffuse[6][0] = 0.0f; colours_diffuse[6][1] = 0.6f; colours_diffuse[6][2] = 0.6f; colours_diffuse[6][3] = 0.6f; // cyan (tranparent)
	colours_diffuse[7][0] = 0.7f; colours_diffuse[7][1] = 0.7f; colours_diffuse[7][2] = 0.7f; colours_diffuse[7][3] = 1.0f; // white
	colours_diffuse[8][0] = 0.2f; colours_diffuse[8][1] = 0.2f; colours_diffuse[8][2] = 0.6f; colours_diffuse[8][3] = 1.0f; // light blue
	colours_diffuse[9][0] = 0.9f; colours_diffuse[9][1] = 0.4f; colours_diffuse[9][2] = 0.7f; colours_diffuse[9][3] = 1.0f; // violett
}

static int vertex_with_max_cell_span = -1;
static Span max_cell_span;

static int cell_with_max_vertex_span = -1;
static Span max_vertex_span;

typedef hash_map<int, Span> my_hash;

static my_hash* cell_span_hash = 0;

static void usage()
{
	fprintf ( stderr,"usage:\n" );
	fprintf ( stderr,"sv_diagram mesh.sva\n" );
	fprintf ( stderr,"sv_diagram -isvc < mesh.svc\n" );
	fprintf ( stderr,"sv_diagram -win 640 480 \n" );
	fprintf ( stderr,"sv_diagram -win 1600 1200 -i mesh.off\n" );
	fprintf ( stderr,"sv_diagram -h\n" );
	fprintf ( stderr,"\n" );
	fprintf ( stderr,"in addition to our formats SVA, SVB, and SVC this tool\n" );
	fprintf ( stderr,"also supports OFF (optionally gzipped).\n" );
}

static void vizBegin()
{
	cell_with_max_vertex_span = -1;
	max_vertex_span.start = 0;
	max_vertex_span.end = 0;
	vertex_with_max_cell_span = -1;
	max_cell_span.start = 0;
	max_cell_span.end = 0;

	if ( svreadopener->file_name )
	{
		fprintf ( stderr,"loading '%s' ...\n", svreadopener->file_name );
	}

	if ( svreader ) // if we have a mesh reader then just re-open the file
	{
		if ( ! ( svreadopener->reopen ( svreader ) ) )
		{
			fprintf ( stderr,"ERROR: cannot reopen svreader\n" );
			usage();
		}
	}
	else
	{
		if ( ! ( svreader = svreadopener->open() ) )
		{
			fprintf ( stderr,"ERROR: cannot open the mesh reader\n" );
			usage();
		}
	}

	if ( svreader->nverts == -1 || svreader->ncells == -1 )
	{
		fprintf ( stderr,"ERROR: no vertex and cell count info ... exiting\n" );
		exit ( 1 );
	}

	nverts = svreader->nverts;
	ncells = svreader->ncells;

	aspect_ratio = ( float ) nverts/ ( float ) ncells;

	fprintf ( stderr,"nverts %d ncells %d\n",nverts,ncells );

	EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
	if ( EVERY_NTH_STEP == 0 )
	{
		EVERY_NTH_STEP = 1;
	}
	TOTAL_STEPS = ( ncells / EVERY_NTH_STEP ) + 1;
	NEXT_STEP = EVERY_NTH_STEP;

	EVERY_NTH_CELL = ncells / EXACTLY_N_CELLS;
	if ( EVERY_NTH_CELL == 0 )
	{
		EVERY_NTH_CELL = 1;
	}
	TOTAL_CELLS = ( ncells / EVERY_NTH_CELL ) + 1;
	NEXT_CELL = 0;

	illustrated_cells_num = 0;

	if ( TOTAL_CELLS > illustrated_cells_alloc )
	{
		illustrated_cells_alloc = TOTAL_CELLS;
		if ( illustrated_cells )
		{
			delete [] illustrated_cells;
		}
		illustrated_cells = new int[illustrated_cells_alloc*10];
	}

	if ( cell_span_hash )
	{
		cell_span_hash->clear();
	}
	else
	{
		cell_span_hash = new my_hash;
	}
}

static void vizEnd()
{
	if ( svreader )
	{
		svreader->close ( true );
		delete svreader;
		svreader = 0;
	}
}

static int vizContinue()
{
	int i;
	Span span;
	SVevent element;
	my_hash::iterator hash_element;

	while ( (element = svreader->read_element())!=SV_EOF )
	{
		if ( element == SV_TETRAHEDRON )
		{
			if ( svreader->c_count >= NEXT_CELL )
			{
				// x-coordinate (cell axis)
				illustrated_cells[illustrated_cells_num*10] = SV_TETRAHEDRON;
				illustrated_cells[illustrated_cells_num*10 + 1] = svreader->c_count;
				for ( i = 0; i < 4; i++ )
				{
					// y-coordinate (vertex axis)
					illustrated_cells[illustrated_cells_num*10+2+i] = svreader->c_idx[i];
				}
				illustrated_cells_num++;
				NEXT_CELL += EVERY_NTH_CELL;

				for ( i = 0; i < 4; i++ )
				{
					hash_element = cell_span_hash->find ( svreader->c_idx[i] );
					if ( hash_element == cell_span_hash->end() )
					{
						span.start = svreader->c_count;
						span.end = svreader->c_count;
						cell_span_hash->insert ( my_hash::value_type ( svreader->c_idx[i], span ) );
					}
					else
					{
						( *hash_element ).second.end = svreader->c_count;
					}
				}
			}
			else
			{
				for ( i = 0; i < 4; i++ )
				{
					hash_element = cell_span_hash->find ( svreader->c_idx[i] );
					if ( hash_element != cell_span_hash->end() )
					{
						( *hash_element ).second.end = svreader->c_count;
					}
				}
			}
		}
		else if ( element == SV_HEXAHEDRON )
		{
			if ( svreader->c_count >= NEXT_CELL )
			{
				// x-coordinate (cell axis)
				illustrated_cells[illustrated_cells_num*10] = SV_HEXAHEDRON;
				illustrated_cells[illustrated_cells_num*10 + 1] = svreader->c_count;
				for ( i = 0; i < 8; i++ )
				{
					// y-coordinate (vertex axis)
					illustrated_cells[illustrated_cells_num*10+2+i] = svreader->c_idx[i];
				}
				illustrated_cells_num++;
				NEXT_CELL += EVERY_NTH_CELL;

				for ( i = 0; i < 8; i++ )
				{
					hash_element = cell_span_hash->find ( svreader->c_idx[i] );
					if ( hash_element == cell_span_hash->end() )
					{
						span.start = svreader->c_count;
						span.end = svreader->c_count;
						cell_span_hash->insert ( my_hash::value_type ( svreader->c_idx[i], span ) );
					}
					else
					{
						( *hash_element ).second.end = svreader->c_count;
					}
				}
			}
			else
			{
				for ( i = 0; i < 8; i++ )
				{
					hash_element = cell_span_hash->find ( svreader->c_idx[i] );
					if ( hash_element != cell_span_hash->end() )
					{
						( *hash_element ).second.end = svreader->c_count;
					}
				}
			}
		}

		if ( svreader->c_count > NEXT_STEP )
		{
			NEXT_STEP += EVERY_NTH_STEP;
			break;
		}
	}

	if ( element )
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

static void myReshape ( int w, int h )
{
	glutReshapeWindow ( WindowW,WindowH );
}

static void myIdle()
{
	if ( AnimationOn )
	{
		AnimationOn = vizContinue();
		if ( !AnimationOn )
		{
			WorkingOn = 0;
			vizEnd();
		}
		glutPostRedisplay();
	}
}

static void myMouseFunc ( int button, int state, int x, int y )
{
	OldX=x;
	OldY=y;
	if ( button == GLUT_LEFT_BUTTON )
	{
		LeftButtonDown = !state;
		MiddleButtonDown = 0;
		RightButtonDown = 0;
	}
	else if ( button == GLUT_RIGHT_BUTTON )
	{
		LeftButtonDown = 0;
		MiddleButtonDown = 0;
		RightButtonDown = !state;
	}
	else if ( button == GLUT_MIDDLE_BUTTON )
	{
		LeftButtonDown = 0;
		MiddleButtonDown = !state;
		RightButtonDown = 0;
	}
}

static void myMotionFunc ( int x, int y )
{
	float RelX = ( x-OldX ) / ( float ) glutGet ( ( GLenum ) GLUT_WINDOW_WIDTH );
	float RelY = ( y-OldY ) / ( float ) glutGet ( ( GLenum ) GLUT_WINDOW_HEIGHT );
	OldX=x;
	OldY=y;
	if ( LeftButtonDown )
	{
		if ( InteractionMode == 0 )
		{
		}
		else if ( InteractionMode == 1 )
		{
			DistX-=RelX;
			DistY+=RelY;
		}
		else if ( InteractionMode == 2 )
		{
			DistZ-=RelY*DistZ;
		}
	}
	else if ( MiddleButtonDown )
	{
		DistX-=RelX*1.0f;
		DistY+=RelY*1.0f;
	}
	glutPostRedisplay();
}

static void MyMenuFunc ( int value );

static void myKeyboard ( unsigned char Key, int x, int y )
{
	switch ( Key )
	{
		case 'Q':
		case 'q':
		case 27:
			exit ( 0 );
			break;
		case ' ': // stationary, translate, or zoom
			if ( InteractionMode == 2 )
			{
				InteractionMode = 1;
			}
			else
			{
				InteractionMode += 1;
			}
			glutPostRedisplay();
			break;
		case '>': //zoom in
			DistZ-=0.1f;
			break;
		case '<': //zoom out
			DistZ+=0.1f;
			break;
		case '-':
			POINTSIZE -= 1.0f;
			if ( POINTSIZE < 1.0f ) POINTSIZE = 1.0f;
			fprintf ( stderr,"POINTSIZE %d\n", ( int ) POINTSIZE );
			glutPostRedisplay();
			break;
		case '=':
			POINTSIZE += 1.0f;
			fprintf ( stderr,"POINTSIZE %d\n", ( int ) POINTSIZE );
			glutPostRedisplay();
			break;
		case '_':
			LINEWIDTH_MAXSPANS -= 1.0f;
			if ( LINEWIDTH_MAXSPANS < 1.0f ) LINEWIDTH_MAXSPANS = 1.0f;
			fprintf ( stderr,"LINEWIDTH_MAXSPANS %d\n", ( int ) LINEWIDTH_MAXSPANS );
			glutPostRedisplay();
			break;
		case '+':
			LINEWIDTH_MAXSPANS += 1.0f;
			fprintf ( stderr,"LINEWIDTH_MAXSPANS %d\n", ( int ) LINEWIDTH_MAXSPANS );
			glutPostRedisplay();
			break;
		case 'T':
		case 't':
			SHOW_CELL_SPAN = !SHOW_CELL_SPAN;
			fprintf ( stderr,"SHOW_CELL_SPAN %d\n",SHOW_CELL_SPAN );
			glutPostRedisplay();
			break;
		case 'O':
		case 'o':
			AnimationOn = 0;
			break;
		case 'V':
		case 'v':
			SHOW_VERTEX_SPAN = !SHOW_VERTEX_SPAN;
			fprintf ( stderr,"SHOW_VERTEX_SPAN %d\n",SHOW_VERTEX_SPAN );
			glutPostRedisplay();
			break;
		case 'M':
		case 'm':
			SHOW_MAX_SPANS = !SHOW_MAX_SPANS;
			fprintf ( stderr,"SHOW_MAX_SPANS %d\n",SHOW_MAX_SPANS );
			glutPostRedisplay();
			break;
		case 'R':
		case 'r':
			SHOW_REFERENCES = !SHOW_REFERENCES;
			fprintf ( stderr,"SHOW_REFERENCES %d\n",SHOW_REFERENCES );
			glutPostRedisplay();
			break;
		case 'D':
		case 'd':
			PrintToFileOn=1;
			if ( Framebuffer ) delete [] Framebuffer;
			Framebuffer = new unsigned char[WindowW*WindowH*3];
			fprintf ( stderr,"print_to_file %d\n",PrintToFileOn );
			break;
		case 'P':
		case 'p':
			AnimationOn = !AnimationOn;
		case 'S':
		case 's':
			if ( WorkingOn == 0 )
			{
				vizBegin();
				WorkingOn = vizContinue();
			}
			else
			{
				WorkingOn = vizContinue();
			}
			if ( WorkingOn == 0 )
			{
				vizEnd();
				AnimationOn = 0;
				PrintToFileOn = 0;
			}
			glutPostRedisplay();
			break;
		case 'K':
		case 'k':
			printf ( "DistX = %ff; DistY = %ff; DistZ = %ff;\n",DistX,DistY,DistZ );
			break;
		case '1':
			MyMenuFunc ( 1 );
			break;
		case '2':
			MyMenuFunc ( 2 );
			break;
		case '3':
			MyMenuFunc ( 3 );
			break;
		case '4':
			MyMenuFunc ( 4 );
			break;
		case '5':
			MyMenuFunc ( 5 );
			break;
		case '6':
			MyMenuFunc ( 6 );
			break;
		case '7':
			MyMenuFunc ( 7 );
			break;
		case '8':
			MyMenuFunc ( 8 );
			break;
		case '9':
			MyMenuFunc ( 9 );
			break;
		case '0':
			MyMenuFunc ( 10 );
			break;
	};
}

static void MyMenuFunc ( int value )
{
	if ( value >= 100 )
	{
		if ( value <= 102 )
		{
			InteractionMode = value - 100;
			glutPostRedisplay();
		}
		else if ( value == 103 )
		{
			myKeyboard ( 's',0,0 );
		}
		else if ( value == 104 )
		{
			myKeyboard ( 'p',0,0 );
		}
		else if ( value == 105 )
		{
			myKeyboard ( 'o',0,0 );
		}
		else if ( value == 109 )
		{
			myKeyboard ( 'q',0,0 );
		}
		else if ( value == 151 )
		{
			myKeyboard ( 'r',0,0 );
		}
		else if ( value == 152 )
		{
			myKeyboard ( 'v',0,0 );
		}
		else if ( value == 153 )
		{
			myKeyboard ( 't',0,0 );
		}
		else if ( value == 154 )
		{
			myKeyboard ( 'm',0,0 );
		}
	}
	else if ( value == 40 )
	{
		EXACTLY_N_STEPS = 5;
	}
	else if ( value == 41 )
	{
		EXACTLY_N_STEPS = 10;
	}
	else if ( value == 42 )
	{
		EXACTLY_N_STEPS = 25;
	}
	else if ( value == 43 )
	{
		EXACTLY_N_STEPS = 50;
	}
	else if ( value == 44 )
	{
		EXACTLY_N_STEPS = 100;
	}
	else if ( value == 45 )
	{
		EXACTLY_N_STEPS = 250;
	}
	else if ( value == 46 )
	{
		EXACTLY_N_STEPS = 500;
	}
	else if ( value == 47 )
	{
		EXACTLY_N_STEPS = 1000;
	}
	else if ( value == 61 )
	{
		EXACTLY_N_CELLS = 50000;
	}
	else if ( value == 62 )
	{
		EXACTLY_N_CELLS = 100000;
	}
	else if ( value == 63 )
	{
		EXACTLY_N_CELLS = 250000;
	}
	else if ( value == 64 )
	{
		EXACTLY_N_CELLS = 500000;
	}
	else if ( value == 65 )
	{
		EXACTLY_N_CELLS = 1000000;
	}
	else if ( value == 66 )
	{
		EXACTLY_N_CELLS = 2000000;
	}
	else if ( value == 67 )
	{
		EXACTLY_N_CELLS = 3000000;
	}
	else if ( value == 68 )
	{
		EXACTLY_N_CELLS = 5000000;
	}
}

static void displayMessage( )
{
	glColor3f ( 0.7f, 0.7f, 0.7f ); // Set colour to grey
	glMatrixMode ( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D ( 0.0f, 1.0f, 0.0f, 1.0f );
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2f ( 0.03f, 0.95f );

	if ( InteractionMode == 0 )
	{
	}
	else if ( InteractionMode == 1 )
	{
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 't' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'r' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'a' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'n' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 's' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'l' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'a' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 't' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'e' );
	}
	else if ( InteractionMode == 2 )
	{
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'z' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'o' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'o' );
		glutBitmapCharacter ( GLUT_BITMAP_8_BY_13, 'm' );
	}

	glPopMatrix();
	glMatrixMode ( GL_PROJECTION );
	glPopMatrix();
}

static void myDisplay()
{
	int sample[3];

	glClear ( GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	glViewport ( 0,0,WindowW,WindowH );

	glMatrixMode ( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective ( 30.0f, ( float ) WindowW/WindowH,0.00000125f,5.0f );

	glMatrixMode ( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt ( DistX,DistY,DistZ, DistX,DistY,0, 0,1,0 );

	glPushMatrix();

	glTranslatef ( -1.0, aspect_ratio, 0.0f );

	if ( ncells && nverts )
	{
		glScalef ( 2.0f/ncells, - ( aspect_ratio*2 ) /nverts, 1.0f );
	}

	// line width for spans

	glLineWidth ( 1.0f );

	// draw vertex spans (grey)

	if ( SHOW_VERTEX_SPAN && illustrated_cells_num )
	{
		// z-coordinate
		sample[2] = 0;

		glColor4fv ( colours_diffuse[7] );
		glBegin ( GL_LINES );
		for ( int i = 0, j = 0; i < illustrated_cells_num; i++,j+=10 )
		{
			int cellType = illustrated_cells[j];
			int numVertices;
			if(cellType==SV_TETRAHEDRON)
				numVertices = 4;
			else if(cellType==SV_HEXAHEDRON)
				numVertices = 8;
			else
			{
				fprintf(stderr, "Error: unsupported element type\n");
				exit(1);
			}
			
			// x-coordinate (cell axis)
			sample[0] = illustrated_cells[j+1];

			// y-coordinate (vertex axis)
			sample[1] = VecMinNiv ( & ( illustrated_cells[j+2] ), numVertices );
			glVertex3iv ( sample );
			sample[1] = VecMaxNiv ( & ( illustrated_cells[j+2] ), numVertices );
			glVertex3iv ( sample );

			if ( SHOW_MAX_SPANS )
			{
				if ( ( sample[1] - illustrated_cells[j+1] ) > ( max_vertex_span.end - max_vertex_span.start ) )
				{
					cell_with_max_vertex_span = sample[0];
					max_vertex_span.start = illustrated_cells[j+1];
					max_vertex_span.end = sample[1];
				}
			}
		}
		glEnd();
	}

	// draw cells spans (green)

	if ( SHOW_CELL_SPAN && cell_span_hash )
	{
		// z-coordinate
		sample[2] = 0;

		glColor4fv ( colours_diffuse[6] );

		glStencilOp ( GL_REPLACE, GL_REPLACE, GL_REPLACE );
		glStencilFunc ( GL_NOTEQUAL, 1, 1 );
		glEnable ( GL_STENCIL_TEST );

		glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		glEnable ( GL_BLEND );

		glBegin ( GL_LINES );
		for ( my_hash::iterator hash_element = cell_span_hash->begin(); hash_element != cell_span_hash->end(); hash_element++ )
		{
			// y-coordinate (vertex axis)
			sample[1] = ( *hash_element ).first;
			// x-coordinate (cell axis)
			sample[0] = ( *hash_element ).second.start;
			// start line
			glVertex3iv ( sample );
			// x-coordinate (cell axis)
			sample[0] = ( *hash_element ).second.end;
			// end line
			glVertex3iv ( sample );

			if ( SHOW_MAX_SPANS )
			{
				if ( ( ( *hash_element ).second.end - ( *hash_element ).second.start ) > ( max_cell_span.end - max_cell_span.start ) )
				{
					vertex_with_max_cell_span = ( *hash_element ).first;
					max_cell_span = ( *hash_element ).second;
				}
			}
		}
		glEnd();

		glDisable ( GL_BLEND );
		glDisable ( GL_STENCIL_TEST );
	}

	// draw the cross references as points (violett)

	if ( SHOW_REFERENCES && illustrated_cells_num )
	{
		// set point size
		glPointSize ( POINTSIZE );

		// z-coordinate
		sample[2] = 0;

		glColor4fv ( colours_diffuse[5] );
		glBegin ( GL_POINTS );
		for ( int i = 0, j = 0; i < illustrated_cells_num; i++, j+=10 )
		{
			int cellType = illustrated_cells[j];
			int numVertices;
			if(cellType==SV_TETRAHEDRON)
				numVertices = 4;
			else if(cellType==SV_HEXAHEDRON)
				numVertices = 8;
			else
			{
				fprintf(stderr, "Error: unsupported element type\n");
				exit(1);
			}
			// x-coordinate (cell axis)
			sample[0] = illustrated_cells[j+1];

			for(int k=0; k< numVertices; ++k)
			{
				// y-coordinate (vertex axis)
				sample[1] = illustrated_cells[j+2+k];
				// draw point
				glVertex3iv ( sample );
			}
		}
		glEnd();
	}

	// draw the max spans (red)

	if ( SHOW_MAX_SPANS && vertex_with_max_cell_span != -1 )
	{
		glLineWidth ( LINEWIDTH_MAXSPANS );
		glColor4fv ( colours_diffuse[1] );
		glBegin ( GL_LINES );
		// y-coordinate (vertex axis)
		sample[1] = vertex_with_max_cell_span;
		// x-coordinate (cell axis)
		sample[0] = max_cell_span.start;
		// start line
		glVertex3iv ( sample );
		// x-coordinate (cell axis)
		sample[0] = max_cell_span.end;
		// end line
		glVertex3iv ( sample );
		glEnd();
		glLineWidth ( 1.0f );
	}

	if ( SHOW_MAX_SPANS && cell_with_max_vertex_span != -1 )
	{
		glLineWidth ( LINEWIDTH_MAXSPANS );
		glColor4fv ( colours_diffuse[1] );
		glBegin ( GL_LINES );
		// x-coordinate (cell axis)
		sample[0] = cell_with_max_vertex_span;
		// y-coordinate (vertex axis)
		sample[1] = max_vertex_span.start;
		// start line
		glVertex3iv ( sample );
		// y-coordinate (vertex axis)
		sample[1] = max_vertex_span.end;
		// end line
		glVertex3iv ( sample );
		glEnd();
		glLineWidth ( 1.0f );
	}

	glPopMatrix();

	glLineWidth ( LINEWIDTH_BOX );
	glColor4fv ( colours_diffuse[0] );
	glBegin ( GL_LINE_LOOP );
	glVertex3f ( -1.0f, aspect_ratio, 0.0f );
	glVertex3f ( 1.0f, aspect_ratio, 0.0f );
	glVertex3f ( 1.0f, -aspect_ratio, 0.0f );
	glVertex3f ( -1.0f, -aspect_ratio, 0.0f );
	glEnd();

	if ( !PrintToFileOn )
	{
		displayMessage();
	}
	else
	{
		char FileName[256], Command[256];
		sprintf ( FileName, "./temp.ppm" );
		glPixelStorei ( GL_PACK_ALIGNMENT, 1 );
#ifdef _WIN32
		glReadPixels ( 0, 0, WindowW, WindowH, GL_RGB, GL_UNSIGNED_BYTE, Framebuffer );
#else
		glReadPixels ( ( 1024-WindowW ) /2, ( 1024-WindowH ) /2, WindowW, WindowH, GL_RGB, GL_UNSIGNED_BYTE, Framebuffer );
#endif
		SavePPM ( FileName, Framebuffer, WindowW, WindowH );
#ifdef _WIN32
		sprintf ( Command, "i_view32.exe temp.ppm /convert=%s%d%d%d%d.jpg",PrintFileName, ( ( Time/1000 ) %10 ), ( ( Time/100 ) %10 ), ( ( Time/10 ) %10 ), ( Time%10 ) );
#else
		sprintf ( Command, "convert ./temp.ppm ./%s%d%d%d%d.jpg",PrintFileName, ( ( Time/1000 ) %10 ), ( ( Time/100 ) %10 ), ( ( Time/10 ) %10 ), ( Time%10 ) );
//      sprintf(Command, "imgcopy -q%d ./temp.ppm ./pics/%s%d%d%d%d.jpg",PrintQuality,PrintFileName,((Time/1000)%10),((Time/100)%10),((Time/10)%10),(Time%10));
#endif
		printf ( "performing: '%s'\n", Command );
		system ( Command );
		Time++;
	}
	glutSwapBuffers();
}

int main ( int argc, char *argv[] )
{
	int i;

	glutInit ( &argc, argv );

	// first look over the command line arguments for input
	for ( i = 1; i < argc; i++ )
	{
		if ( strncmp ( argv[i],"-i",2 ) == 0 )
		{
			if ( svreadopener == 0 )
			{
				// apparently there is a point reader to be opened
				svreadopener = new SVreadOpener();
			}
			if ( strcmp ( argv[i],"-i" ) == 0 ) // the commandline '-i' followed by the input filename
			{
				i++;
				svreadopener->set_file_name ( argv[i] );
			}
			else // the commandline '-ixxx' specifies the input format 'xxx'
			{
				svreadopener->set_file_format ( & ( argv[i][2] ) );
			}
		}
	}
	// second look over the command line arguments for options
	for ( i = 1; i < argc; i++ )
	{
		if ( strncmp ( argv[i],"-i",2 ) == 0 )
		{
			if ( strcmp ( argv[i],"-i" ) == 0 ) i++;
			// already processed
		}
		else if ( strcmp ( argv[1],"-h" ) == 0 || strcmp ( argv[1],"-help" ) == 0 )
		{
			usage();
		}
		else if ( strcmp ( argv[i],"-win" ) == 0 )
		{
			i++;
			WindowW = atoi ( argv[i] );
			i++;
			WindowH = atoi ( argv[i] );
		}
		else if ( strcmp ( argv[i],"-every" ) == 0 )
		{
			i++;
			EVERY_NTH_STEP = atoi ( argv[i] );
		}
		else if ( strcmp ( argv[i],"-steps" ) == 0 )
		{
			i++;
			EXACTLY_N_STEPS = atoi ( argv[i] );
		}
		else if ( strcmp ( argv[i],"-samples" ) == 0 )
		{
			i++;
			EXACTLY_N_CELLS = atoi ( argv[i] );
		}
		else if ( ( i == argc-1 ) && svreadopener == 0 )
		{
			// last argument unknown and no input specified yet ... try it
			svreadopener = new SVreadOpener();
			svreadopener->set_file_name ( argv[i] );
		}
		else
		{
			usage();
		}
	}

	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_STENCIL );
	glutInitWindowSize ( WindowW,WindowH );
	glutInitWindowPosition ( 10,100 );
	glutCreateWindow ( "Layout Diagram for Tetrahedral Volume Meshes" );

	glClearColor ( 1.0f, 1.0f, 1.0f, 1.0f );
	glClearStencil ( 0 );

	InitColors();

	glutDisplayFunc ( myDisplay );
	glutReshapeFunc ( myReshape );
	glutIdleFunc ( myIdle );

	glutMouseFunc ( myMouseFunc );
	glutMotionFunc ( myMotionFunc );
	glutKeyboardFunc ( myKeyboard );

	// steps sub menu
	int menuSteps = glutCreateMenu ( MyMenuFunc );
	glutAddMenuEntry ( "in 5 steps", 40 );
	glutAddMenuEntry ( "in 10 steps", 41 );
	glutAddMenuEntry ( "in 25 steps", 42 );
	glutAddMenuEntry ( "in 50 steps", 43 );
	glutAddMenuEntry ( "in 100 steps", 44 );
	glutAddMenuEntry ( "in 250 steps", 45 );
	glutAddMenuEntry ( "in 500 steps", 46 );
	glutAddMenuEntry ( "in 1000 steps", 47 );

	// cell sub menu
	int menuCells = glutCreateMenu ( MyMenuFunc );
	glutAddMenuEntry ( "50,000 cells", 61 );
	glutAddMenuEntry ( "100,000 cells", 62 );
	glutAddMenuEntry ( "250,000 cells", 63 );
	glutAddMenuEntry ( "500,000 cells", 64 );
	glutAddMenuEntry ( "1,000,000 cells", 65 );
	glutAddMenuEntry ( "2,000,000 cells", 66 );
	glutAddMenuEntry ( "3,000,000 cells", 67 );
	glutAddMenuEntry ( "5,000,000 cells", 68 );

	// main menu
	glutCreateMenu ( MyMenuFunc );
	glutAddMenuEntry ( "translate <SPACE>", 101 );
	glutAddMenuEntry ( "zoom <SPACE>", 102 );
	glutAddMenuEntry ( "", 0 );
	glutAddMenuEntry ( "<s>tep", 103 );
	glutAddMenuEntry ( "<p>lay", 104 );
	glutAddMenuEntry ( "st<o>p", 105 );
	glutAddMenuEntry ( "", 0 );
	glutAddMenuEntry ( "show <r>eferences (+/-)", 151 );
	glutAddMenuEntry ( "show <v>-span", 152 );
	glutAddMenuEntry ( "show <t>-span", 153 );
	glutAddMenuEntry ( "show <m>ax", 154 );
	glutAddMenuEntry ( "", 0 );
	glutAddSubMenu ( "steps ...", menuSteps );
	glutAddSubMenu ( "sampling ...", menuCells );
	glutAddMenuEntry ( "", 0 );
	glutAddMenuEntry ( "<Q>UIT", 109 );
	glutAttachMenu ( GLUT_RIGHT_BUTTON );

	glutMainLoop();

	return 0;
}
